from rclpy.node import Node
from geometry_msgs.msg import Twist
from std_msgs.msg import Float32

from . import CanBus
from .drive_controller import AckermanController
from .hardware_interfaces import Motor, Steering, Brake, BrakeFeedback, SpeedSensor


import can
import rclpy


class ControlManager(Node):
    def __init__(self):
        super().__init__('control_manager_node')

        # Create the controller
        self.controller = AckermanController(
            wheel_base=1.09, front_track=0.9, back_track=0.98
        )

        # Subscribe to the /cmd_vel topic
        self.cmd_vel_sub = self.create_subscription(
            Twist,  # Message Type
            '/cmd_vel',  # Topic Name
            self.cmd_vel_callback,  # Callback function
            10  # QoS Profile
        )

        self.speed_publisher = self.create_publisher(
            Float32,
            '/kart_speed',
            10
        )

        # Get the ROS2 logger
        logger = self.get_logger()

        # Define the hardware interfaces
        motor = Motor(logger)
        steering = Steering(logger)

        brake_fb = BrakeFeedback(logger)
        brake = Brake(logger, brake_feedback=brake_fb)

        # Speed and Steering sensors
        speed_sensor = SpeedSensor(
            logger,
            speed_publisher=self.speed_publisher
        )

        # Add additional sensors
        motor.add_speed_sensor(speed_sensor)
        motor.add_brake_feedback(brake_fb)

        # Register can bus listeners and notifiers
        can_bus = CanBus()
        can.Notifier(can_bus.bus, [speed_sensor, brake_fb])

        # Set the hardware interfaces in the controller
        self.controller.set_motor(motor)
        self.controller.set_steering(steering)
        self.controller.set_brake(brake)

    def cmd_vel_callback(self, msg: Twist):
        # Prcess the Twist message and change the motor speeds, steering or braking accordingly
        linear_x = msg.linear.x
        angular_z = msg.angular.z

        # Process the linear and angular velocities and set the motor speeds, steering and braking
        steering_angle, speed = self.controller.process_velocity(
            linear_x=linear_x,
            angular_z=angular_z
        )

        self.get_logger().info(
            f"Steering Angle: {steering_angle}, Speed: {speed}"
        )


def main(args=None):
    rclpy.init(args=args)
    control_manager = ControlManager()
    rclpy.spin(control_manager)
    control_manager.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
