from ..hardware_interfaces import Motor, Steering, Brake

from math import atan, cos


class AckermanController(object):
    """
    An simple Ackerman controller for a vehicle with front steering and back wheel drive.
    (the equations are simplified, and represent a bycicle model more than a real vehicle model)
    """

    def __init__(self, wheel_base: float, front_track: float, back_track: float) -> None:
        self._motor: Motor = None
        self._steering: Steering = None
        self._brake: Brake = None

        # The distance between the front and back wheels
        self.WHEEL_BASE = wheel_base
        # The distance between the front wheels
        self.FRONT_TRACK = front_track
        # The distance between the back wheels
        self.BACK_TRACK = back_track

    def set_motor(self, motor: Motor) -> None:
        self._motor = motor

    def set_steering(self, steering: Steering) -> None:
        self._steering = steering

    def set_brake(self, brake: Brake) -> None:
        self._brake = brake

    def process_velocity(self, linear_x: float, angular_z: float) -> tuple[float, float]:
        """
        Process the linear and angular velocities and set the motor speeds, steering and braking

        :param linear_x: The linear velocity of the vehicle
        :param angular_z: The angular velocity of the vehicle

        :return: The front steering angle and the back wheel speed
        """

        # Process the linear and angular velocities and set the motor speeds, steering and braking
        front_steering_angle = self._calculate_front_steering_angle(
            linear_x, angular_z
        )
        back_wheel_speed = self._calculate_back_wheel_speed(
            linear_x, front_steering_angle
        )

        # Set the brake
        # self._brake.set_brake(linear_x == 0)

        # Multiply the steering angle to -1 to turn the wheels in the right direction
        self._steering.set_angle(front_steering_angle * -1)
        self._motor.set_speed(back_wheel_speed)

        return front_steering_angle, back_wheel_speed

    def _calculate_front_steering_angle(self, linear_x: float, angular_z: float) -> float:
        """
        Calculate the front steering angle based on the angular and linear velocities

        :param angular_z: The angular velocity of the vehicle
        :param linear_x: The linear velocity of the vehicle

        :return: The front steering angle in radians
        """

        try:
            # Calculate the front steering angle
            return round(atan(self.WHEEL_BASE * angular_z / linear_x), 2)
        except ZeroDivisionError:
            # If the linear velocity is zero and angular velocity is not 0,
            # set the steering angle to the maximum
            if angular_z != 0.0:
                return self._steering.MAX_STEERING_ANGLE_RAD if angular_z > 0 else self._steering.MIN_STEERING_ANGLE_RAD  # noqa: E501
            else:
                return 0.0

    def _calculate_back_wheel_speed(self, linear_x: float, steering_angle: float) -> float:
        """
        Calculate the back wheel speed based on the linear velocity and the steering angle

        :param linear_x: The linear velocity of the vehicle
        :param steering_angle: The front steering angle in radians

        :return: The back wheel speed in m/s
        """

        # Calculate the back wheel speed
        return round(linear_x / cos(steering_angle), 2)
