from .interface import CanBusInterface
from .speed_sensor import SpeedSensor
from .brake_feedback import BrakeFeedback

from simple_pid import PID

import threading


class Motor(CanBusInterface):
    MIN_THROTTLE = 0
    MAX_THROTTLE = 100
    START_THROTTLE = 20

    MODE1_MAX_SPEED = 1.38889  # m/s
    MODE2_MAX_SPEED = 4.16667  # m/s
    MODE3_MAX_SPEED = 8.33333  # m/s
    MODE4_MAX_SPEED = 16.6667  # m/s

    def __init__(self, logger, mode: float = MODE2_MAX_SPEED) -> None:
        super().__init__(logger, arbitration_id=0x330)
        self.logger.info("Motor Interface Initialized")

        self._mode_max_speed: float = mode
        self._current_target_speed: float = None

        # PID controller constants
        self._Kp: float = 8.0
        self._Ki: float = 6.0
        self._Kd: float = 4.0

        # Initialize the PID controller
        pid = PID(
            self._Kp, self._Ki, self._Kd,
            setpoint=0, starting_output=self.START_THROTTLE
        )
        pid.output_limits = (self.MIN_THROTTLE, self.MAX_THROTTLE)
        pid.sample_time = 0.1
        pid.set_auto_mode(False)

        self._pid: PID = pid

        # Define the adustment thread
        self._adjustment_thread = threading.Thread(
            target=self.adjust_speed, args=(0, 0)
        )
        self._is_thread_running = False

        # Speed sensor
        self._speed_sensor: SpeedSensor = None
        # Brake feedback
        self._brake_feedback: BrakeFeedback = None

    def set_speed(self, speed: float) -> None:
        """
        Set the speed of the vehicle

        :param speed: The speed in m/s
        """

        # If the speed is the same as the current target speed, do nothing
        if speed == self._current_target_speed:
            return

        self._current_target_speed = speed

        # 1 (forward)
        gear = 1
        if speed < 0:
            # 2 (reverse)
            gear = 2
        elif speed == 0:
            # 0 (neutral)
            gear = 0

        # Clamp the speed to the maximum speed
        speed = min(abs(speed), self._mode_max_speed)

        # If the adjusted thread is running, stop it
        if self._adjustment_thread.is_alive():
            # self.logger.info("Stopping the adjustment thread")
            self._is_thread_running = False
            # Wait for the thread to finish
            self._adjustment_thread.join()

        # self.logger.info(f"Setting speed to: {speed}")

        # Set the new arguments for the adjustment thread
        self._adjustment_thread = threading.Thread(
            target=self.adjust_speed, args=(speed, gear)
        )
        self._is_thread_running = True
        # Start the adjustment thread
        self._adjustment_thread.start()

    def adjust_speed(self, speed: float, gear: int) -> None:
        self.logger.info(
            f"Starting adjust thread with target speed: {speed}"
        )

        # Current speed of the vehicle (to avoid logging the same speed multiple times)
        current_speed = -1
        current_throttle = -1

        # Enable and set the target speed of the PID controller
        self._adjust_pid_target(speed)

        # While the speed is not equal to the setpoint
        while abs(self._speed_sensor.speed - self._pid.setpoint) != 0.0:
            if not self._is_thread_running:
                break

            # Log only if the speed is different from the previous speed
            if current_speed != self._speed_sensor.speed:
                self.logger.info(
                    f"Current speed: {self._speed_sensor.speed} | target speed: {self._pid.setpoint}"
                )
                current_speed = self._speed_sensor.speed

            # Feed the current speed to the PID controller
            throttle = int(self._pid(self._speed_sensor.speed))

            # Convert the throttle to the bus data format
            # Format: [throttle, 0, gear, 0, 0, 0, 0, 0]
            data = [throttle, 0, gear, 0, 0, 0, 0, 0]

            # Log only if the throttle is different from the previous throttle
            if current_throttle != throttle:
                self.logger.info(
                    f"Setting throttle to: {throttle} | and gear to: {gear}\n"
                )
                current_throttle = throttle

            # If the brake is engaged, do not send the throttle
            if self._brake_feedback.is_not_braking():
                # Set the speed of the motor
                self.set_bus_data(data)
                self.send_bus_message()

        self.logger.info(
            f"Finished adjust speed thread with target speed: {speed}\n"
        )

        self._pid.set_auto_mode(False)

    def _adjust_pid_target(self, speed: float) -> None:
        start_throttle = self.START_THROTTLE

        # If the car is stopped, start with a specific throttle
        if self._speed_sensor.speed == 0:
            if speed > 0 and speed <= 0.5:
                start_throttle = 20
            elif speed > 0.5 and speed <= 1:
                start_throttle = 25
            elif speed > 1 and speed <= 2:
                start_throttle = 30
            elif speed > 2:
                start_throttle = 40
            else:
                start_throttle = 0

        self._pid.set_auto_mode(True, last_output=start_throttle)

        # Change the setpoint of the PID controller
        self._pid.setpoint = speed

    def add_speed_sensor(self, speed_sensor: SpeedSensor) -> None:
        self._speed_sensor = speed_sensor

    def add_brake_feedback(self, brake_feedback: BrakeFeedback) -> None:
        self._brake_feedback = brake_feedback
