from .interface import CanBusInterface


class Steering(CanBusInterface):
    # The minimum and maximum throttle values the servo motor accepts
    STEERING_MIN_THROTTLE = -1.25
    STEERING_MAX_THROTTLE = 1.25

    # The minimum and maximum steering angle (in radians)
    # (assumed if the turning diameter is 5.75 meters)
    MAX_STEERING_ANGLE_RAD = 0.42  # 24 degrees
    MIN_STEERING_ANGLE_RAD = -0.42  # -24 degrees

    def __init__(self, logger) -> None:
        super().__init__(logger, arbitration_id=0x220)
        self.logger.info("Steering Interface Initialized")

        self._current_angle: float = None

    def set_angle(self, angle: float) -> None:
        """
        Set the steering angle of the vehicle

        :param angle: The angle in radians
        """

        # If the angle is the same as the current angle, do nothing
        if angle == self._current_angle:
            return

        self._current_angle = angle

        # Clamp the angle between the minimum and the maximum steering angle
        angle = min(
            max(angle, Steering.MIN_STEERING_ANGLE_RAD),
            Steering.MAX_STEERING_ANGLE_RAD
        )

        self.logger.info(f"Setting angle to: {angle}")

        # Mathematically assume the throttle based on the received angle
        throttle = angle * Steering.STEERING_MAX_THROTTLE / Steering.MAX_STEERING_ANGLE_RAD
        # Approximate the throttle to the closest 2 decimal places float
        throttle = round(throttle, 2)
        # Make sure the throttle is within the accepted range
        throttle = min(
            max(throttle, Steering.STEERING_MIN_THROTTLE),
            Steering.STEERING_MAX_THROTTLE
        )

        # Convert the throttle to little-endian bytes
        throttle_bytes = CanBusInterface.convert_to_bytes(
            throttle, byteorder='little'
        )
        # Data format: [throttle, throttle, throttle, throttle, 0, 0, 0, 0]
        data = list(throttle_bytes) + [0, 0, 0, 0]

        self.logger.info(f"Setting angle throttle to: {throttle}")
        self.logger.info(f"Setting angle data to: {data}")

        # Set the data and send the message
        self.set_bus_data(data)
        self.send_bus_message()
