from pynput import keyboard

from ..utils.camera_utils import _threshold_lanes, _img_prep
from core.utils import display_image_safe

import threading
import cv2 as cv
import numpy as np
import json
import os


class CalibrationTools(object):
    # Current Camera
    CURRENT_CAMERA = {
        1: 'front',
        2: 'left',
        3: 'right'
    }
    FILE_PATH = os.path.join(
        os.path.dirname(__file__),
        'bev_values.json'
    )
    # Reference Source values for left/right cam based on Kart 1/2  
    SRC_REF_MATRIX = {
        'left': {
            1: np.float32([[610, 110], [848, 220], [100, 70], [848, 540]]),
            2: np.float32([[710, 110], [848, 180], [100, 120], [848, 480]]),
        },
        'right': {
            1: np.float32([[0, 250], [110, 190], [0, 480], [848, 105]]),
            2: np.float32([[0, 220], [115, 160], [0, 480], [848, 83]]),
        },
    }
    # Current car. Values 1 or 2 !!!!!
    KART = 1
    #---------------------------------

    def __init__(self):
        # Initial coordinates (Front Camera)
        self.top_l_x, self.top_l_y = 320, 400
        self.top_r_x, self.top_r_y = 470, 400
        self.bottom_l_x, self.bottom_l_y = 320, 500
        self.bottom_r_x, self.bottom_r_y = 400, 600
        self.front_cam, self.left_cam, self.right_cam = None, None, None
        self.camera_num = 1
        self.is_calibration_running = True

        # Print instructions
        print("""
            ----------------Controls-----------------
            |    Q    |    W    |    E    |    R    |
            |---------|---------|---------|---------|
            | Top_L_x | TOP_R_X | BOT_L_X | BOT_R_X |
            |         |         |         |         |
            |    A    |     S   |     D   |    F    |
            | TOP_L_Y | TOP_R_Y | BOT_L_Y | BOT_R_Y |
            -----------------------------------------
            Shift + Key -> Increase | Key -> Decrease
            M -> Save to file
        """)

        # Initialize values from JSON
        self.initialise_values_json()
        # Create keyboard listener
        # self.listen_to_keyboard()

        # Create and start threads
        self._keyboard_thread = threading.Thread(
            target=self.listen_to_keyboard
        )
        self._keyboard_thread.start()

    def calibrate_all_cameras(self):
        # Display the image
        if self.is_calibration_running is True:
            # Get image from current camera being calibrated
            load = self.get_image_obj()
            gray = _img_prep(load)

            bev = self.to_birds_eye_view(gray)
            # thresh = _threshold_lanes(bev)

            display_image_safe(bev, "Bird's Eye View")
            # display_image_safe("Threshold", thresh)
        else:
            # Stop the calibration
            self._keyboard_thread.join()
            exit()

    def write_to_json_file(
        self,
        top_l: tuple[int, int], top_r: tuple[int, int],
        bottom_l: tuple[int, int], bottom_r: tuple[int, int],
        camera_key: str
    ) -> None:
        dst_data = {
            'top_l': top_l,
            'top_r': top_r,
            'bottom_l': bottom_l,
            'bottom_r': bottom_r
        }

        try:
            with open(self.FILE_PATH, "r+") as file:
                # Load existing data
                data = json.load(file)
                # Update or add data
                data[camera_key] = dst_data
                # Move cursor to the beginning for overwriting
                file.seek(0)
                # Write updated data
                json.dump(data, file, indent=4)
        except Exception as e:
            print(f"An error occurred while writing to the file: {e}")

    def get_image_obj(self) -> cv.Mat | None:
        if self.camera_num > 3:
            self.is_calibration_running = False

        if self.CURRENT_CAMERA[self.camera_num] == 'front':
            load = self.front_cam
        elif self.CURRENT_CAMERA[self.camera_num] == 'left':
            load = self.left_cam
        elif self.CURRENT_CAMERA[self.camera_num] == 'right':
            load = self.right_cam
        else:
            # Handle no Image here
            self.is_calibration_running = False

        return load

    def initialise_values_json(self) -> None:
        """
        Read values from the JSON file and initialise the values in the script's fields
        """
        if self.camera_num > 3:
            self.is_calibration_running = False
            return

        camera_key = self.CURRENT_CAMERA[self.camera_num]
        try:
            with open(self.FILE_PATH, "r") as file:
                # Load existing data
                data = json.load(file)

                # Check for valid direction
                if camera_key not in ["left", "right", "front"]:
                    return None

                # Return values for the specified direction
                transform_matrix = data.get(camera_key)
                if transform_matrix is None:
                    return None

                self.top_l_x = transform_matrix['top_l'][0]
                self.top_l_y = transform_matrix['top_l'][1]

                self.top_r_x = transform_matrix['top_r'][0]
                self.top_r_y = transform_matrix['top_r'][1]

                self.bottom_l_x = transform_matrix['bottom_l'][0]
                self.bottom_l_y = transform_matrix['bottom_l'][1]

                self.bottom_r_x = transform_matrix['bottom_r'][0]
                self.bottom_r_y = transform_matrix['bottom_r'][1]
        except FileNotFoundError:
            print(f"Error: File '{self.FILE_PATH}' not found.")
            return None

    def to_birds_eye_view(self, image: cv.UMat) -> cv.UMat:
        # width, height of output image
        image_size = (640, 800)
        # Change based on which camera is it
        if self.camera_num > 3:
            self.is_calibration_running = False

        if self.CURRENT_CAMERA[self.camera_num] == 'front':
            src = np.float32([[432, 249], [627, 249], [432, 267], [576, 295]])
        elif self.CURRENT_CAMERA[self.camera_num] == 'right':
            src = self.SRC_REF_MATRIX['right'][self.KART]
        elif self.CURRENT_CAMERA[self.camera_num] == 'left':
            src = self.SRC_REF_MATRIX['left'][self.KART]

        # Each point from src get shifted to following coord
        dst = np.float32(
            [[self.top_l_x, self.top_l_y],
             [self.top_r_x, self.top_r_y],
             [self.bottom_l_x, self.bottom_l_y],
             [self.bottom_r_x, self.bottom_r_y]]
        )

        matrix = cv.getPerspectiveTransform(src, dst)
        warped_img = cv.warpPerspective(image, matrix, image_size)

        return warped_img

    def on_press(self, key):
        STEP = 1

        try:
            # Top left X
            if key.char == 'q':
                self.top_l_x = self.top_l_x - STEP
            elif key.char == 'Q':
                self.top_l_x = self.top_l_x + STEP
            # Top left Y
            elif key.char == 'a':
                self.top_l_y = self.top_l_y - STEP
            elif key.char == 'A':
                self.top_l_y = self.top_l_y + STEP

            # Top right X
            elif key.char == "w":
                self.top_r_x = self.top_r_x - STEP
            elif key.char == "W":
                self.top_r_x = self.top_r_x + STEP
            # Top right y
            elif key.char == "s":
                self.top_r_y = self.top_r_y - STEP
            elif key.char == "S":
                self.top_r_y = self.top_r_y + STEP

            # Bottom left X
            elif key.char == "e":
                self.bottom_l_x = self.bottom_l_x - STEP
            elif key.char == "E":
                self.bottom_l_x = self.bottom_l_x + STEP
            # Bottom left Y
            elif key.char == "d":
                self.bottom_l_y = self.bottom_l_y - STEP
            elif key.char == "D":
                self.bottom_l_y = self.bottom_l_y + STEP

            # Bottom right X
            elif key.char == "r":
                self.bottom_r_x = self.bottom_r_x - STEP
            elif key.char == "R":
                self.bottom_r_x = self.bottom_r_x + STEP
            # Bottom right Y
            elif key.char == "f":
                self.bottom_r_y = self.bottom_r_y - STEP
            elif key.char == "F":
                self.bottom_r_y = self.bottom_r_y + STEP

            # Save
            elif key.char == 'm' or key.char == 'M':
                if self.camera_num > 3:
                    return
                # TODO: Change dict value retrieval to .get() method
                print(
                    'Save to Json for Camera: ',
                    self.CURRENT_CAMERA[self.camera_num]
                )

                self.write_to_json_file(
                    (int(self.top_l_x), int(self.top_l_y)),
                    (int(self.top_r_x), int(self.top_r_y)),
                    (int(self.bottom_l_x), int(self.bottom_l_y)),
                    (int(self.bottom_r_x), int(self.bottom_r_y)),
                    self.CURRENT_CAMERA[self.camera_num]
                )
                # Update the camera view
                self.camera_num = self.camera_num + 1
                self.initialise_values_json()

            print(
                self.top_l_x, self.top_l_y, "<--Top-->", self.top_r_x, self.top_r_y
            )
            print(
                self.bottom_l_x, self.bottom_l_y, "<--Bottom-->", self.bottom_r_x, self.bottom_r_y
            )
        except AttributeError:
            print('Exception, unexpected key: ', key)

    def on_release(self, key):
        # Stop listener
        if key == keyboard.Key.esc:
            self.is_calibration_running = False
            return False  # Stop listener

    def listen_to_keyboard(self):
        # Collect events until released
        listener = keyboard.Listener(
            on_press=self.on_press, on_release=self.on_release
        )
        listener.start()  # start to listen on a separate thread
        listener.join()  # remove if main thread is polling self.keys

    # def run_calibration(self):
    #     # Print instructions
    #     print("----------------Controls-----------------")
    #     print("|    Q    |    W    |    E    |    R    |")
    #     print("|---------|---------|---------|---------|")
    #     print("| Top_L_x | TOP_R_X | BOT_L_X | BOT_R_X |")
    #     print("|         |         |         |         |")
    #     print("|    A    |     S   |     D   |    F    |")
    #     print("| TOP_L_Y | TOP_R_Y | BOT_L_Y | BOT_R_Y |")
    #     print("-----------------------------------------")
    #     print("Shift + Key -> Increase | Key -> Decrease")
    #     print("M -> Save to file")

    #     # Initialize values from JSON
    #     self.initialise_values_json()

    #     # Create and start threads
    #     image_thread = threading.Thread(target=self.update_image)
    #     keyboard_thread = threading.Thread(target=self.listen_to_keyboard)

    #     image_thread.start()
    #     keyboard_thread.start()

    #     image_thread.join()
    #     keyboard_thread.join()
