from tkinter import filedialog, Canvas, Toplevel, messagebox
from PIL import Image, ImageTk
from rclpy.node import Node
from sensor_msgs.msg import Image as ROSImage
from cv_bridge import CvBridge

import rclpy
import os
import json
import cv2
import numpy as np
import threading
import tkinter as tk


class BirdEyeViewApp(Node):
    RESIZED_IMAGE_WIDTH = 848
    RESIZED_IMAGE_HEIGHT = 480

    FILE_PATH = os.path.join(
        os.path.dirname(__file__),
        'bev_values.json'
    )

    def __init__(self, root):
        super().__init__('bird_eye_view_app')
        self.root = root
        self.root.title("Bird's Eye View Adjuster")

        self.canvas = Canvas(root, width=800, height=600)
        self.canvas.pack()

        self.button_frame = tk.Frame(root)
        self.button_frame.pack()

        self.upload_button = tk.Button(
            self.button_frame, text="Upload Image", command=self.upload_image)
        self.upload_button.pack(side=tk.LEFT)

        self.reset_button = tk.Button(
            self.button_frame, text="Reset", command=self.reset)
        self.reset_button.pack(side=tk.LEFT)

        self.transform_button = tk.Button(
            self.button_frame, text="Transform", command=self.transform_image)
        self.transform_button.pack(side=tk.LEFT)

        self.calibrate_button = tk.Button(
            self.button_frame, text="Calibrate Camera", command=self.calibrate_camera)
        self.calibrate_button.pack(side=tk.LEFT)

        self.json_button = tk.Button(
            self.button_frame, text="Save to JSON", command=self.save_to_json)
        self.json_button.pack(side=tk.LEFT)

        self.ros_frame = tk.Frame(root)
        self.ros_frame.pack()

        self.topic_entry_label = tk.Label(self.ros_frame, text="ROS2 Topic:")
        self.topic_entry_label.pack(side=tk.LEFT)

        self.topic_entry = tk.Entry(self.ros_frame, width=30)
        self.topic_entry.insert(0, "/camera_front/image_raw")
        self.topic_entry.pack(side=tk.LEFT)

        self.ros_button = tk.Button(
            self.ros_frame, text="Subscribe", command=self.subscribe_topic)
        self.ros_button.pack(side=tk.LEFT)

        self.coords_label = tk.Label(root, text="Coordinates: ")
        self.coords_label.pack()

        self.coords = []
        self.image = None
        self.photo_image = None
        self.resized_image = None
        self.resize_ratio = None
        self.bridge = CvBridge()

        self.entries_frame = tk.Frame(root)
        self.entries_frame.pack()

        self.entries = []
        self.original_coords_labels = []
        for i in range(4):
            x_label = tk.Label(self.entries_frame, text=f"Point {i + 1} X:")
            x_label.grid(row=i, column=0)
            x_entry = tk.Entry(self.entries_frame, width=5)
            x_entry.grid(row=i, column=1)
            x_entry.bind(
                "<Up>", lambda e, idx=i,
                axis='x': self.adjust_coordinate(idx, axis, 1)
            )
            x_entry.bind(
                "<Down>", lambda e, idx=i,
                axis='x': self.adjust_coordinate(idx, axis, -1)
            )

            y_label = tk.Label(self.entries_frame, text=f"Point {i + 1} Y:")
            y_label.grid(row=i, column=2)
            y_entry = tk.Entry(self.entries_frame, width=5)
            y_entry.grid(row=i, column=3)
            y_entry.bind(
                "<Up>", lambda e, idx=i,
                axis='y': self.adjust_coordinate(idx, axis, 1)
            )
            y_entry.bind(
                "<Down>", lambda e, idx=i,
                axis='y': self.adjust_coordinate(idx, axis, -1)
            )

            self.entries.append((x_entry, y_entry))

            original_coords_label = tk.Label(
                self.entries_frame, text="Original: (0, 0)"
            )
            original_coords_label.grid(row=i, column=4)
            self.original_coords_labels.append(original_coords_label)

        self.update_button = tk.Button(
            self.entries_frame, text="Update Coordinates", command=self.update_coords
        )
        self.update_button.grid(row=4, columnspan=5, pady=10)

        self.subscription = None
        self.image_received = False

        self.dragging_point_index = None
        self.top = None
        self.canvas_warped = None

        self.calibration_images = []
        # Standard chessboard size for calibration
        self.chessboard_size = (7, 4)

        # Source transform matrix coordinates
        self.src_transform_points = None

        # Which camera is being calibrated now
        self.current_camera = 'front'

    def upload_image(self):
        file_path = filedialog.askopenfilename()
        if file_path:
            self.image = Image.open(file_path)
            self.resized_image = self.resize_image(
                self.image, self.RESIZED_IMAGE_WIDTH, self.RESIZED_IMAGE_HEIGHT
            )
            self.photo_image = ImageTk.PhotoImage(self.resized_image)
            self.canvas.create_image(
                0, 0, anchor=tk.NW, image=self.photo_image
            )
            self.canvas.bind("<Button-1>", self.on_click)
            self.canvas.bind("<B1-Motion>", self.drag_point)
            self.canvas.bind("<ButtonRelease-1>", self.release_point)

    def resize_image(self, image, max_width, max_height):
        width_ratio = max_width / image.width
        height_ratio = max_height / image.height
        self.resize_ratio = min(width_ratio, height_ratio)
        new_width = int(image.width * self.resize_ratio)
        new_height = int(image.height * self.resize_ratio)
        return image.resize((new_width, new_height), Image.ANTIALIAS)

    def on_click(self, event):
        if len(self.coords) < 4:
            self.get_coords(event)
        else:
            self.dragging_point_index = self.find_nearest_point(
                event.x, event.y
            )

    def get_coords(self, event):
        if len(self.coords) < 4:
            x, y = event.x, event.y
            self.coords.append((x, y))
            self.canvas.create_oval(x - 3, y - 3, x + 3, y + 3, fill='red')
            if len(self.coords) == 4:
                self.display_coords()
                self.update_entries()

    def display_coords(self):
        coords_text = "Coordinates: " + \
            ', '.join([f"({x}, {y})" for x, y in self.coords])
        self.coords_label.config(text=coords_text)

    def reset(self):
        self.coords = []
        self.coords_label.config(text="Coordinates: ")
        self.canvas.delete("all")
        if self.resized_image:
            self.photo_image = ImageTk.PhotoImage(self.resized_image)
            self.canvas.create_image(
                0, 0, anchor=tk.NW, image=self.photo_image
            )
        for x_entry, y_entry in self.entries:
            x_entry.delete(0, tk.END)
            y_entry.delete(0, tk.END)
        for label in self.original_coords_labels:
            label.config(text="Original: (0, 0)")

    def update_entries(self):
        for i, (x_entry, y_entry) in enumerate(self.entries):
            x_entry.delete(0, tk.END)
            y_entry.delete(0, tk.END)
            x_entry.insert(0, self.coords[i][0])
            y_entry.insert(0, self.coords[i][1])
        self.update_original_coords_labels()

    def update_coords(self):
        new_coords = []
        for x_entry, y_entry in self.entries:
            try:
                x = int(x_entry.get())
                y = int(y_entry.get())
                new_coords.append((x, y))
            except ValueError:
                continue
        if len(new_coords) == 4:
            self.coords = new_coords
            self.display_coords()
            self.update_canvas()
            self.transform_image()
            self.update_original_coords_labels()

    def update_canvas(self):
        self.canvas.delete("all")
        if self.resized_image:
            self.photo_image = ImageTk.PhotoImage(self.resized_image)
            self.canvas.create_image(
                0, 0, anchor=tk.NW, image=self.photo_image
            )
        for x, y in self.coords:
            self.canvas.create_oval(x - 3, y - 3, x + 3, y + 3, fill='red')

    def update_original_coords_labels(self):
        if len(self.coords) == 4:
            original_coords = [
                (x / self.resize_ratio, y / self.resize_ratio) for x, y in self.coords
            ]
            for i, (x, y) in enumerate(original_coords):
                self.original_coords_labels[i].config(
                    text=f"Original: ({int(x)}, {int(y)})")

    def adjust_coordinate(self, index, axis, delta):
        x_entry, y_entry = self.entries[index]
        if axis == 'x':
            entry = x_entry
        else:
            entry = y_entry

        try:
            value = int(entry.get())
            value += delta
            entry.delete(0, tk.END)
            entry.insert(0, value)
            self.update_coords()
        except ValueError:
            pass

    def subscribe_topic(self):
        topic_name = self.topic_entry.get()
        if topic_name:
            if self.subscription is not None:
                self.destroy_subscription(self.subscription)

            self.subscription = self.create_subscription(
                ROSImage,
                topic_name,
                self.image_callback,
                10)
            self.image_received = False

    def image_callback(self, msg):
        if self.image_received:
            return

        self.image_received = True
        image_cv = self.bridge.imgmsg_to_cv2(msg, desired_encoding='bgr8')
        self.image = Image.fromarray(cv2.cvtColor(image_cv, cv2.COLOR_BGR2RGB))
        self.resized_image = self.resize_image(
            self.image, self.RESIZED_IMAGE_WIDTH, self.RESIZED_IMAGE_HEIGHT
        )
        self.photo_image = ImageTk.PhotoImage(self.resized_image)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.photo_image)
        self.canvas.bind("<Button-1>", self.on_click)
        self.canvas.bind("<B1-Motion>", self.drag_point)
        self.canvas.bind("<ButtonRelease-1>", self.release_point)

    def transform_image(self):
        if len(self.coords) == 4:
            original_coords = [
                (x / self.resize_ratio, y / self.resize_ratio) for x, y in self.coords
            ]
            pts = np.array(original_coords, dtype="float32")
            (tl, tr, bl, br) = pts

            # Save transform points
            self.src_transform_points = pts

            # print(pts)

            dst = np.array([
                [0, 0],
                [640 - 1, 0],
                [0, 800 - 1],
                [640 - 1, 800 - 1]
            ], dtype="float32")

            image_cv = np.array(self.image)
            M = cv2.getPerspectiveTransform(pts, dst)
            warped = cv2.warpPerspective(image_cv, M, (640, 800))

            warped_image = Image.fromarray(warped)
            warped_photo = ImageTk.PhotoImage(warped_image)

            if self.top is None or not self.top.winfo_exists():
                self.top = Toplevel(self.root)
                self.top.title("Bird's Eye View")
                self.canvas_warped = Canvas(self.top, width=640, height=800)
                self.canvas_warped.pack()

            self.canvas_warped.create_image(
                0, 0, anchor=tk.NW, image=warped_photo
            )
            # Keep a reference to avoid garbage collection
            self.canvas_warped.image = warped_photo

    def drag_point(self, event):
        if self.dragging_point_index is not None:
            self.coords[self.dragging_point_index] = (event.x, event.y)
            self.update_canvas()
            self.update_entries()
            self.transform_image()

    def release_point(self, event):
        self.dragging_point_index = None

    def find_nearest_point(self, x, y):
        nearest_index = None
        min_dist = float('inf')
        for i, (px, py) in enumerate(self.coords):
            dist = (x - px) ** 2 + (y - py) ** 2
            if dist < min_dist:
                min_dist = dist
                nearest_index = i

        return nearest_index

    def calibrate_camera(self):
        if not self.image_received:
            messagebox.showerror("Error", "No image received for calibration.")
            return

        self.calibration_images.append(self.image)
        if len(self.calibration_images) >= 10:  # Assuming we need at least 10 images
            ret, camera_matrix, dist_coeffs, rvecs, tvecs, debug_images = self.perform_calibration()
            if ret:
                messagebox.showinfo(
                    "Calibration Result", f"Camera Matrix: \n{camera_matrix}\n\nDistortion Coefficients: \n{dist_coeffs}"
                )
                self.calibration_images = []  # Reset after successful calibration
            else:
                messagebox.showerror(
                    "Error", "Camera calibration failed. Check the debug images for more information."
                )
                self.display_debug_images(debug_images)
        else:
            messagebox.showinfo(
                "Info", f"Captured {len(self.calibration_images)} images for calibration. Need at least 10."
            )

    def perform_calibration(self):
        objp = np.zeros(
            (self.chessboard_size[0] * self.chessboard_size[1], 3), np.float32
        )
        objp[:, :2] = np.mgrid[
            0:self.chessboard_size[0],
            0:self.chessboard_size[1]
        ].T.reshape(-1, 2)

        objpoints = []  # 3d point in real world space
        imgpoints = []  # 2d points in image plane.

        debug_images = []

        for image in self.calibration_images:
            image_cv = np.array(image)
            gray = cv2.cvtColor(image_cv, cv2.COLOR_BGR2GRAY)
            ret, corners = cv2.findChessboardCorners(
                gray, self.chessboard_size, None
            )

            if ret:
                objpoints.append(objp)
                imgpoints.append(corners)
                debug_image = image_cv.copy()
                cv2.drawChessboardCorners(
                    debug_image, self.chessboard_size, corners, ret
                )
                debug_images.append(debug_image)
            else:
                debug_images.append(image_cv)

        if objpoints and imgpoints:
            ret, camera_matrix, dist_coeffs, rvecs, tvecs = cv2.calibrateCamera(
                objpoints, imgpoints, gray.shape[::-1], None, None
            )
            return ret, camera_matrix, dist_coeffs, rvecs, tvecs, debug_images
        else:
            return False, None, None, None, None, debug_images

    def display_debug_images(self, images):
        if not hasattr(self, 'debug_window') or not self.debug_window.winfo_exists():
            self.debug_window = Toplevel(self.root)
            self.debug_window.title("Debug Images")
            self.debug_canvas = Canvas(
                self.debug_window, width=800, height=600
            )
            self.debug_canvas.pack()

        for i, image in enumerate(images):
            debug_image = Image.fromarray(
                cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            )
            debug_photo = ImageTk.PhotoImage(debug_image)
            self.debug_canvas.create_image(
                0, i * 600, anchor=tk.NW, image=debug_photo
            )
            # Keep a reference to avoid garbage collection
            self.debug_canvas.image = debug_photo

    def undistort_camera(self, img: cv2.UMat, mtx, dist):
        undistorted_img = cv2.undistort(img, mtx, dist, None, mtx)
        return undistorted_img

    def save_to_json(self) -> None:
        # Get current camera
        camera_key = str(self.topic_entry.get())
        try:
            # Get camera based on topic path -> looks like /camera_front/image_raw
            camera_key = camera_key.split('/')[1]
            camera_key = camera_key.split('_')[1]
        except IndexError as e:
            print('Incorrect topic name')
            return None

        # Floor all the src_transform_points and convert to int
        if self.coords is not None:
            self.coords = np.floor(self.coords)
            self.coords = self.coords.astype(int)

        src_data = {
            'top_l': [int(self.coords[0][0]), int(self.coords[0][1])],
            'top_r': [int(self.coords[1][0]), int(self.coords[1][1])],
            'bottom_l': [int(self.coords[2][0]), int(self.coords[2][1])],
            'bottom_r': [int(self.coords[3][0]), int(self.coords[3][1])]
        }

        # print(src_data)

        # Save to correct camera
        try:
            with open(self.FILE_PATH, "r+") as file:
                # Load existing data
                data = json.load(file)

                # Update or add data
                if self.coords is not None:
                    data[camera_key] = src_data
                    # Move cursor to the beginning for overwriting
                    file.seek(0)
                    # Write updated data
                    json.dump(data, file, indent=4)
        except Exception as e:
            print(f"An error occurred while writing to the file: {e}")


def main(args=None):
    rclpy.init(args=args)

    root = tk.Tk()
    app = BirdEyeViewApp(root)

    def ros_spin():
        rclpy.spin(app)

    spin_thread = threading.Thread(target=ros_spin, daemon=True)
    spin_thread.start()

    root.mainloop()

    app.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()
