from .utils import DetectedLaneResult
from .utils import _get_lane_dist_from_image, _get_single_lane_distance, _detect_grass_percentage
from ..utils import display_image_safe
from .utils import BEV_MATRIX_FRONT, perspective_transforms_left, perspective_transforms_right
from dotenv import load_dotenv

import cv2
import numpy as np
import os
import json

# When a horizontal line at this dist, only use side cams
HORIZONTAL_DIST_CUTOFF = 200
GRASS_THRESHOLD = 20  # % of grass to be found in side cam frame to turn the other way
MAX_DIST_TO_LANE = 500  # in px. Max distance to a lane that Node can return


def detect(front_image: cv2.Mat, right_image: cv2.Mat | None, left_image: cv2.Mat | None) -> DetectedLaneResult:
    # Try to get distances using front camera
    result, right_front_track_x = _get_lane_dist_from_image(front_image)

    # CHeck that both distances are found
    right_lane_dist = result.get('right', {}).get('dist')
    left_lane_dist = result.get('left', {}).get('dist')
    hor_line_dist = result.get('horizontal', {}).get('dist')

    if hor_line_dist is None:
        hor_line_dist = 99999

    # Use right camera data if right lane is not visible in the front camera view
    if (right_lane_dist is None or hor_line_dist <= HORIZONTAL_DIST_CUTOFF) and right_image is not None:
        # use right camera image detection
        right_cam_result = _get_single_lane_distance(right_image, False)
        if right_cam_result is not None:
            result['right']['dist'] = right_cam_result

        # If no lane found with side cam check for grass
        if right_cam_result is None:
            right_grass_percent = _detect_grass_percentage(right_image, False)
            if right_grass_percent >= GRASS_THRESHOLD:
                result['right']['dist'] = None
            else:
                result['right']['dist'] = right_cam_result

    # Use left camera data if left lane is not visible in the front camera view
    if (left_lane_dist is None or hor_line_dist <= HORIZONTAL_DIST_CUTOFF) and left_image is not None:
        # use left camera image detection
        left_cam_result = _get_single_lane_distance(left_image, True)
        if left_cam_result is not None:
            result['left']['dist'] = left_cam_result

        # If no lane found with side cam check for grass
        if left_cam_result is None:
            left_grass_percent = _detect_grass_percentage(left_image, True)
            if left_grass_percent >= GRASS_THRESHOLD:
                result['left']['dist'] = None
            else:
                result['left']['dist'] = left_cam_result

    # Amplify the distance if one lane is detected and the other is not
    right_lane_dist = result.get('right', {}).get('dist')
    left_lane_dist = result.get('left', {}).get('dist')

    if left_lane_dist is not None and left_lane_dist <= 30 and right_lane_dist is None:
        result['left']['dist'] += 200

    if right_lane_dist is not None and right_lane_dist <= 30 and left_lane_dist is None:
        result['right']['dist'] += 200

    # Make sure that the values are not larger than the frame (500px)
    right_lane_dist = result.get('right', {}).get('dist')
    left_lane_dist = result.get('left', {}).get('dist')

    if left_lane_dist is not None and left_lane_dist > MAX_DIST_TO_LANE:
        result['left']['dist'] = MAX_DIST_TO_LANE

    if right_lane_dist is not None and right_lane_dist > MAX_DIST_TO_LANE:
        result['right']['dist'] = MAX_DIST_TO_LANE

    return result, right_front_track_x


IS_ZEBRA_CORSSING_THRESHOLD = 4


def is_zebra_crossing(img: cv2.Mat) -> tuple[bool, int]:
    if img is None:
        return False, 0

    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    blurred = cv2.GaussianBlur(gray, (15, 15), 6)

    _, thresh = cv2.threshold(
        blurred,
        180, 255,
        cv2.THRESH_BINARY
    )

    contours, _ = cv2.findContours(
        thresh.copy(),
        cv2.RETR_TREE,
        cv2.CHAIN_APPROX_SIMPLE
    )

    valid_contours = 0
    for c in contours:
        # if the contour is not sufficiently large, ignore it
        if cv2.contourArea(c) < 3000:
            continue

        valid_contours += 1

        # get the min area rect
        rect = cv2.minAreaRect(c)
        box = cv2.boxPoints(rect)
        # convert all coordinates floating point values to int
        box = np.int0(box)
        # draw a red 'nghien' rectangle
        cv2.drawContours(img, [box], 0, (0, 0, 255), 2)

        # display_image_safe(img, 'ZebraCrossing - contours')

    return valid_contours >= IS_ZEBRA_CORSSING_THRESHOLD, valid_contours


# Load BEV values from json file based on the simulation or real world
dotenv_path = os.path.join(
    os.path.dirname(__file__),
    '.env'
)

load_dotenv(dotenv_path=dotenv_path)

BEV_VALUES_FILE = os.path.join(
    os.path.dirname(__file__),
    'calibration/bev_values.json'
)

if os.getenv('GAZEBO_SIMULATION', 'false') == 'true':
    BEV_VALUES_FILE = os.path.join(
        os.path.dirname(__file__),
        'calibration/bev_values_gazebo.json'
    )


def parse_bev_values() -> None:
    """
    Gets BEV transformation matrix values from json file after calibration sequence.
    """
    FILE_PATH = BEV_VALUES_FILE

    def make_np_array(matrix_inp: dict):
        top_l_x = transform_matrix['top_l'][0]
        top_l_y = transform_matrix['top_l'][1]

        top_r_x = transform_matrix['top_r'][0]
        top_r_y = transform_matrix['top_r'][1]

        bottom_l_x = transform_matrix['bottom_l'][0]
        bottom_l_y = transform_matrix['bottom_l'][1]

        bottom_r_x = transform_matrix['bottom_r'][0]
        bottom_r_y = transform_matrix['bottom_r'][1]

        dst = np.float32(
            [[top_l_x, top_l_y],
             [top_r_x, top_r_y],
             [bottom_l_x, bottom_l_y],
             [bottom_r_x, bottom_r_y]]
        )
        return dst

    try:
        with open(FILE_PATH, "r") as file:
            # Load existing data
            data = json.load(file)

            # Update values for Front camera
            transform_matrix = data.get('front')
            if transform_matrix is None:
                return None
            else:
                BEV_MATRIX_FRONT['src'] = make_np_array(transform_matrix)

            # Update values for Left camera
            transform_matrix = data.get('left')
            if transform_matrix is None:
                return None
            else:
                perspective_transforms_left[1]['src'] = make_np_array(
                    transform_matrix)
                perspective_transforms_left[2]['src'] = make_np_array(
                    transform_matrix)

            # Update values for Right camera
            transform_matrix = data.get('right')
            if transform_matrix is None:
                return None
            else:
                perspective_transforms_right[1]['src'] = make_np_array(
                    transform_matrix)
                perspective_transforms_right[2]['src'] = make_np_array(
                    transform_matrix)

    except FileNotFoundError:
        print(f"Error: File '{FILE_PATH}' not found.")
        return None
