from typing import Callable, Optional

from simple_pid import PID
from rclpy.clock import Clock
from geometry_msgs.msg import Twist
from rclpy.impl.rcutils_logger import RcutilsLogger

from .utils import DetectedLaneResult
from ..utils import StopKartException


class LanePlanner(object):
    MAX_LINEAR_VELOCITY = round(20 / 3.6, 2)  # hm/h -> m/s
    MAX_ANGULAR_VELOCITY = 1.0  # Positive is left, negative is right

    PID_MIN_OUTPUT = -50
    PID_MAX_OUTPUT = 50
    # Must be negative to match the angular velocity direction
    PID_DEVISION_FACTOR = -100

    PID_INPUT_DIFFERENCE_THRESHOLD = 20
    PID_INPUT_COUNTER_THRESHOLD = 2

    SIDE_LANE_STORED_DETECTIONS = 10  # detections number
    # maximum difference between the last detection and the new detection to be considered
    SIDE_LANE_DETECTION_THRESHOLD = 50  # pixels
    # number of detections to ignore
    SIDE_LANE_IGNORE_THRESHOLD = 2  # detections number
    # value for the kart movement depending to the ditance to the opposite lane
    SIDE_LANE_THRESHOLD = 80  # pixels

    HORIZONTAL_LANE_STOP_THRESHOLD = 200  # pixels
    HORIZONTAL_LANE_STOP_TIMER_THRESHOLD = 5  # seconds
    HORIZONTAL_LANE_DETECTION_COUNT_THRESHOLD = 5  # detections number

    def __init__(self, logger: RcutilsLogger, clock: Clock, get_can_start: Callable[[], bool]) -> None:
        self._logger = logger
        self._clock = clock
        self.get_can_start = get_can_start

        self._twist_message = Twist()
        self._linear_velocity: float = self.MAX_LINEAR_VELOCITY
        self._angular_velocity: float = 0.0

        # Store detections
        # Left
        self._left_detections: list[float] = []
        self._last_left_detection = None
        self._left_ignore_count: int = 0
        self._left_none_count: int = 0
        # Right
        self._right_detections: list[float] = []
        self._last_right_detection = None
        self._right_ignore_count: int = 0
        self._right_none_count: int = 0
        # Front
        self._horizontal_last_detected_time = clock.now()
        self._was_horizontal_detected = False
        self._horizontal_lane_type = None
        self._horizontal_last_lane_type = None
        self._horizontal_lane_detection_count = 0

        # PID controller constants
        self.Kp = 1.8
        self.Ki = 1.3
        self.Kd = 0.8
        self.Kc = 3.5  # curvature constant

        self._last_pid_input = 0.0
        self._last_pid_input_counter = 0

        # Initialize the PID controller
        self.pid = PID(self.Kp, self.Ki, self.Kd, setpoint=0)
        self.pid.output_limits = (self.PID_MIN_OUTPUT, self.PID_MAX_OUTPUT)
        self.pid.sample_time = 0.05

    def get_horizontal_lane_type(self) -> Optional[str]:
        return self._horizontal_lane_type

    def plan(self, lane_detections: DetectedLaneResult) -> Twist:
        """
        Plan the navigation based on the lane detections

        :param lane_detections: DetectedLaneResult
        :return: Twist
        :raises StopKartException: Stop the kart if there is a horizontal lane detected
        """
        # Check for the horizontal lane
        self._detect_horizontal_lane(lane_detections)

        # Compare the distances of the left and right lanes
        left_lane_dist = lane_detections.get('left', {}).get('dist')
        right_lane_dist = lane_detections.get('right', {}).get('dist')

        if left_lane_dist is not None and right_lane_dist is not None:
            # Store the detections
            left_lane_dist, right_lane_dist = self._store_and_process_detections(
                left_lane_dist, right_lane_dist
            )

            curvature = self._estimate_curvature(lane_detections)

            # Determine the angular velocity based on the lane distances
            pid_input = round((left_lane_dist - right_lane_dist) / 10, 2)

            # Check for the input to not have a big difference
            if (
                abs(pid_input - self._last_pid_input) > self.PID_INPUT_DIFFERENCE_THRESHOLD and
                self._last_pid_input_counter < self.PID_INPUT_COUNTER_THRESHOLD
            ):
                pid_input = self._last_pid_input
                self._last_pid_input_counter += 1
            else:
                self._last_pid_input = pid_input
                self._last_pid_input_counter = 0

            # Perform the PID calculation
            pid_result = float(
                self.pid(pid_input) + (self.Kc * curvature)
            )
            pid_result = round(pid_result / self.PID_DEVISION_FACTOR, 2)

            # Set the velocities based on the pid output and the curvature
            self._set_velocities(pid_result, curvature)

            # Logs for debugging
            # self._logger.info(f'Curvature: {curvature}')
            # self._logger.info(f'PID Input: {pid_input}')
            # self._logger.info(f'PID Result: {pid_result}')
        else:
            self._none_lane_processing(lane_detections)

        # Debug log ================================================================================================================
        # self._logger.info(
        #     f"""
        #         Far:     {lane_detections.get('far_left', {}).get('dist')} <-> {lane_detections.get('far_right', {}).get('dist')}
        #         Current: {lane_detections.get('left', {}).get('dist')} <-> {lane_detections.get('right', {}).get('dist')}
        #         Linear: {self._linear_velocity}
        #         Angular: {self._angular_velocity}
        #     """
        # )
        # ==========================================================================================================================

        if not self.get_can_start():
            # Reset the velocities
            self._linear_velocity = 0.0
            self._angular_velocity = 0.0
            # Reset the PID controller
            self.pid.reset()

        # Create a PriorityTwist message
        return self._get_twist_message()

    def _get_twist_message(self) -> Twist:
        self._twist_message.linear.x = round(self._linear_velocity, 2)
        self._twist_message.angular.z = round(self._angular_velocity, 2)

        return self._twist_message

    def _detect_horizontal_lane(self, lane_detections: DetectedLaneResult) -> None:
        horizontal_lane = lane_detections.get(
            'horizontal', {}
        ).get('type', 'none')

        if horizontal_lane is not None and horizontal_lane.lower() in ('zebra', 'horizontal'):
            # Stop the kart if a horizontal lane is detected
            horizontal_dist = lane_detections.get('horizontal', {}).get('dist')

            # Check for the lane detection count
            if self._horizontal_last_lane_type == horizontal_lane.lower():
                self._horizontal_lane_detection_count += 1
            else:
                self._horizontal_lane_detection_count = 0

            # Save the lane type
            self._horizontal_last_lane_type = horizontal_lane.lower()

            # Check for the lane detection count
            if (
                self._horizontal_lane_detection_count >= self.HORIZONTAL_LANE_DETECTION_COUNT_THRESHOLD and
                self.get_horizontal_lane_type() != horizontal_lane.lower()
            ):
                self._horizontal_lane_type = horizontal_lane.lower()
                # Debug log
                self._logger.info(
                    f'Horizontal lane detected: {horizontal_lane.lower()}'
                )

            # The horizontal lane is detected and is close enough
            if horizontal_dist <= self.HORIZONTAL_LANE_STOP_THRESHOLD:
                # Stop the kart if the horizontal lane is detected the first time
                if not self._was_horizontal_detected:
                    self._was_horizontal_detected = True
                    raise StopKartException(horizontal_lane.lower())
                else:
                    # If the kart did not detect any other lanes for 5 seconds
                    if (
                        self._clock.now() - self._horizontal_last_detected_time
                    ).nanoseconds >= self.HORIZONTAL_LANE_STOP_TIMER_THRESHOLD * 1e9:  # seconds to nanoseconds
                        self._was_horizontal_detected = False

                    # Reset the timer
                    self._horizontal_last_detected_time = self._clock.now()
        else:
            self._was_horizontal_detected = False
            self._horizontal_lane_type = None

    def _set_velocities(self, pid_result: float, curvature: float) -> None:
        self._angular_velocity = pid_result
        self._linear_velocity = self.MAX_LINEAR_VELOCITY

        # If the curvature is too high, reduce the linear velocity
        if abs(curvature) >= 0.7:
            self._linear_velocity = 0.7
        else:
            self._linear_velocity = self.MAX_LINEAR_VELOCITY

        # Check for an horizontal lane and slow down the kart
        if self._was_horizontal_detected:
            self._linear_velocity = 0.5

    def _none_lane_processing(self, lane_detections: DetectedLaneResult) -> None:
        left_lane_dist = lane_detections.get('left', {}).get('dist')
        right_lane_dist = lane_detections.get('right', {}).get('dist')
        left_far_lane_dist = lane_detections.get('far_left', {}).get('dist')
        right_far_lane_dist = lane_detections.get('far_right', {}).get('dist')

        if right_lane_dist is None and self._right_none_count > self.SIDE_LANE_IGNORE_THRESHOLD:
            self._right_none_count == 0
            # If the right lane is None, and the distance to the left lane is smaller than N pixels, turn right (the kart is too close to the left lane)
            if (
                (left_lane_dist is not None and left_lane_dist <= self.SIDE_LANE_THRESHOLD) and
                (left_far_lane_dist is None or left_far_lane_dist <= 0)
            ):
                self._angular_velocity = -self.MAX_ANGULAR_VELOCITY
            else:
                # If the right lane is None, but the distance to the left lane is greater than N pixels, turn left (the kart is too far from the left lane)
                self._angular_velocity = self.MAX_ANGULAR_VELOCITY

                # Use the far distance to proof check the kart must turn left
                if right_far_lane_dist is not None and right_far_lane_dist > 0:
                    self._angular_velocity = -self.MAX_ANGULAR_VELOCITY

            self._linear_velocity = 0.5
        else:
            self._right_none_count += 1

        if left_lane_dist is None and self._left_none_count > self.SIDE_LANE_IGNORE_THRESHOLD:
            self._left_none_count == 0
            # If the left lane is None, and the distance to the right lane is smaller than N pixels, turn left (the kart is too close to the right lane)
            if (
                (right_lane_dist is not None and right_lane_dist <= self.SIDE_LANE_THRESHOLD) and
                (right_far_lane_dist is None or right_far_lane_dist <= 0)
            ):
                self._angular_velocity = self.MAX_ANGULAR_VELOCITY
            else:
                # If the left lane is None, but the distance to the right lane is greater than N pixels, turn right (the kart is too far from the right lane)
                self._angular_velocity = -self.MAX_ANGULAR_VELOCITY

                # Use the far distance to proof check the kart must turn right
                if left_far_lane_dist is not None and left_far_lane_dist > 0:
                    self._angular_velocity = self.MAX_ANGULAR_VELOCITY

            self._linear_velocity = 0.5
        else:
            self._left_none_count += 1

        if (
            left_lane_dist is None and self._left_none_count > self.SIDE_LANE_IGNORE_THRESHOLD and
            right_lane_dist is None and self._right_none_count > self.SIDE_LANE_IGNORE_THRESHOLD
        ):
            # Stop if no lanes are detected
            self._linear_velocity = 0.0
            self._angular_velocity = 0.0

    def _store_and_process_detections(self, left_lane_dist: float, right_lane_dist: float) -> tuple[float, float]:
        # Store the last detection regardless of the distance
        self._last_left_detection = left_lane_dist
        self._last_right_detection = right_lane_dist

        store_left = False
        store_right = False

        # Check for the detection threshold
        if len(self._left_detections) > 0 and abs(left_lane_dist - self._left_detections[-1]) <= self.SIDE_LANE_DETECTION_THRESHOLD:
            self._left_ignore_count += 1
        else:
            self._left_ignore_count = 0
            store_left = True

        if len(self._right_detections) > 0 and abs(right_lane_dist - self._right_detections[-1]) <= self.SIDE_LANE_DETECTION_THRESHOLD:
            self._right_ignore_count += 1
        else:
            self._right_ignore_count = 0
            store_right = True

        # Check for the ignore threshold
        if self._left_ignore_count > self.SIDE_LANE_IGNORE_THRESHOLD:
            store_left = True

        if self._right_ignore_count > self.SIDE_LANE_IGNORE_THRESHOLD:
            store_right = True

        if store_left:
            # Store the last N detections
            if len(self._left_detections) > self.SIDE_LANE_STORED_DETECTIONS:
                self._left_detections.pop(0)

            self._left_detections.append(left_lane_dist)

        if store_right:
            # Store the last N detections
            if len(self._right_detections) > self.SIDE_LANE_STORED_DETECTIONS:
                self._right_detections.pop(0)

            self._right_detections.append(right_lane_dist)

        return (self._left_detections[-1], self._right_detections[-1])

    def _estimate_curvature(self, lane_result: DetectedLaneResult) -> float:
        far_left_dist = lane_result['far_left']['dist']
        far_right_dist = lane_result['far_right']['dist']

        if far_left_dist is None:
            far_left_dist = 0

        if far_right_dist is None:
            far_right_dist = 0

        try:
            curvature = (
                (far_right_dist - far_left_dist) /
                (far_left_dist + far_right_dist)
            ) * -1
        except ZeroDivisionError:
            curvature = 0.0

        return round(curvature, 2)
