from typing import Optional

from rclpy.node import Node
from cv_bridge import CvBridge
from std_msgs.msg import String, Bool
from std_srvs.srv import SetBool
from sensor_msgs.msg import Image
from geometry_msgs.msg import Twist
from std_msgs.msg import Float32
from std_srvs.srv import SetBool
from cv_bridge import CvBridge

from custom_interfaces.msg import PriorityTwist, Lanes, Lane
from ..utils import StopKartException
from .lane_planning import LanePlanner
from .lane_detection import DetectedLaneResult, detect, is_zebra_crossing, parse_bev_values

from ..utils import display_image_safe

import rclpy


class LaneTrackingNode(Node):
    def __init__(self):
        super().__init__('lane_tracking_node')
        self.get_logger().info('LaneTrackingNode started')

        # Subscribe to the /camera_front/image_raw topic
        self.create_subscription(
            Image,
            '/camera_front/image_raw',
            self._image_front_callback,
            10
        )
        self._front_image: Optional[Image] = None

        # Subscribe to the /camera_left/image_raw topic
        self.create_subscription(
            Image,
            '/camera_left/image_raw',
            self._image_left_callback,
            10
        )
        self._left_image: Optional[Image] = None

        # Subscribe to the /camera_right/image_raw topic
        self.create_subscription(
            Image,
            '/camera_right/image_raw',
            self._image_right_callback,
            10
        )
        self._right_image: Optional[Image] = None

        # Publish to the /lane_x_position topic
        self._lane_x_position_publisher = self.create_publisher(
            Float32,
            '/lane_x_position',
            10
        )
        self._lane_x_position_message: Float32 = Float32()

        # Publish to the /lane_detections topic
        self._lane_detections_publisher = self.create_publisher(
            Lanes,
            '/lane_detections',
            10
        )
        self._lane_detections_message: Lanes = Lanes()
        self._lane_detections_message.left = Lane()
        self._lane_detections_message.right = Lane()
        self._lane_detections_message.far_left = Lane()
        self._lane_detections_message.far_right = Lane()
        self._lane_detections_message.horizontal = Lane()

        # Publish to the /horizontal_line_detect topic
        self._horizontal_line_publisher = self.create_publisher(
            String,
            '/horizontal_line_detect',
            10
        )

        # Publish to the /priority_cmd_vel topic
        self._cmd_vel_publisher = self.create_publisher(
            PriorityTwist,
            '/priority_cmd_vel',
            10
        )
        self._priority_twist_message: PriorityTwist = PriorityTwist()

        # Subscribe to the /can_start topic
        self.create_subscription(
            Bool,
            '/can_start',
            self._can_start_callback,
            10
        )
        self._can_start: bool = False

        # Service to stop the kart
        self._client = self.create_client(SetBool, 'set_can_start')
        while not self._client.wait_for_service(timeout_sec=1.0):
            self.get_logger().info('service not available, waiting again...')

        self._stop_request = SetBool.Request()

        # Initialize the OpenCV bridge
        self._cv_bridge = CvBridge()

        # Get the BEV values from calibration bev_values.json
        parse_bev_values()

        # Initialize the lane planner
        self._planner = LanePlanner(
            logger=self.get_logger(),
            clock=self.get_clock(),
            get_can_start=self._get_can_start,
        )

    def _image_front_callback(self, msg: Image) -> None:
        # Process the Image message
        self._front_image = self._cv_bridge.imgmsg_to_cv2(msg)

        # Perform lane detection
        detected_lanes, right_lane_x_track = detect(
            front_image=self._front_image,
            right_image=self._right_image,
            left_image=self._left_image
        )

        # Publish the detected lanes
        self._publish_detected_lanes(detected_lanes, right_lane_x_track)
        # Process the detected lanes
        self.process_detected_lanes(detected_lanes)

    def _image_left_callback(self, msg: Image) -> None:
        # Process the Image message
        self._left_image = self._cv_bridge.imgmsg_to_cv2(msg)

        # Testing =============================================================
        # from .utils.left_camera_utils import _left_side_cam_birds_eye_view
        # from .utils.camera_utils import _img_prep

        # test_img = _img_prep(self._left_image)
        # res = _left_side_cam_birds_eye_view(test_img)
        # display_image_safe(res, 'Left Side Cam Birds Eye View')
        # ======================================================================

    def _image_right_callback(self, msg: Image) -> None:
        # Process the Image message
        self._right_image = self._cv_bridge.imgmsg_to_cv2(msg)

        # Testing =============================================================
        # from .utils.right_camera_utils import _right_side_cam_birds_eye_view
        # from .utils.camera_utils import _img_prep

        # test_img = _img_prep(self._right_image)
        # res = _right_side_cam_birds_eye_view(test_img)
        # display_image_safe(res, 'RIght Side Cam Birds Eye View')
        # ======================================================================

    def _can_start_callback(self, msg: Bool) -> None:
        if self._get_can_start() != msg.data:
            self.get_logger().info(f'Can start: {msg.data}')
            self._can_start = msg.data

    def _get_can_start(self) -> bool:
        return self._can_start

    def process_detected_lanes(self, lane_detection: DetectedLaneResult) -> None:  # noqa: E501
        try:
            twist_message = self._planner.plan(lane_detection)
        except StopKartException:
            # Stop the kart if a horizontal lane is detected
            self.send_kart_stop_request()
            return

        horizontal_lane = self._planner.get_horizontal_lane_type()
        if horizontal_lane is not None:
            is_zebra = horizontal_lane == 'zebra'

            if not is_zebra:
                is_zebra, detected_contours = is_zebra_crossing(
                    self._front_image
                )
                # self.get_logger().info(
                #     f'ZebraPad: {is_zebra} - Contours: {detected_contours}'
                # )

            if is_zebra:
                self._horizontal_line_publisher.publish(String(data='zebra'))

        self.publish_velocity(twist_message)

    def _publish_detected_lanes(self, lane_detection: DetectedLaneResult, right_lane_x_track: Optional[float]) -> None:
        # Publish the tracked lane x position
        if right_lane_x_track is not None:
            self._lane_x_position_message.data = right_lane_x_track
            self._lane_x_position_publisher.publish(
                self._lane_x_position_message
            )

        # Publish the detected lanes
        # Left
        left_dist = lane_detection.get('left', {}).get('dist')
        self._lane_detections_message.left.dist = float(
            left_dist if left_dist is not None else -999
        )
        self._lane_detections_message.left.type = str(
            lane_detection.get('left', {}).get('type')
        )

        # Right
        right_dist = lane_detection.get('right', {}).get('dist')
        self._lane_detections_message.right.dist = float(
            right_dist if right_dist is not None else -999
        )
        self._lane_detections_message.right.type = str(
            lane_detection.get('right', {}).get('type')
        )

        # Far left
        far_left_dist = lane_detection.get('far_left', {}).get('dist')
        self._lane_detections_message.far_left.dist = float(
            far_left_dist if far_left_dist is not None else -999
        )
        self._lane_detections_message.far_left.type = str(
            lane_detection.get('far_left', {}).get('type')
        )

        # Far right
        far_right_dist = lane_detection.get('far_right', {}).get('dist')
        self._lane_detections_message.far_right.dist = float(
            far_right_dist if far_right_dist is not None else -999
        )
        self._lane_detections_message.far_right.type = str(
            lane_detection.get('far_right', {}).get('type')
        )

        # Horizontal
        horizontal_dist = lane_detection.get('horizontal', {}).get('dist')
        self._lane_detections_message.horizontal.dist = float(
            horizontal_dist if horizontal_dist is not None else -999
        )
        self._lane_detections_message.horizontal.type = str(
            lane_detection.get('horizontal', {}).get('type')
        )

        # Publish the detected lanes
        self._lane_detections_publisher.publish(self._lane_detections_message)

    def publish_velocity(self, twist_message: Twist, priority: int = 5) -> None:
        self._priority_twist_message.priority = priority
        self._priority_twist_message.twist = twist_message
        self._cmd_vel_publisher.publish(self._priority_twist_message)

    def send_kart_stop_request(self) -> None:
        self._stop_request.data = False
        self._future = self._client.call_async(self._stop_request)


def main(args=None):
    rclpy.init(args=args)
    node = LaneTrackingNode()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
