from typing import TypedDict, Optional

from .front_camera_utils import _to_birds_eye_view, _track_lanes
from .right_camera_utils import _right_side_cam_birds_eye_view, _get_right_distance_to_lane
from .right_camera_utils import _get_right_distance_to_lanes_coordinate
from .left_camera_utils import _left_side_cam_birds_eye_view, _get_left_distance_to_lanes_coordinate, _get_left_distance_to_lanes

# To display an image safely, use the 'display_image_safe' function
from core.utils.custom_cv2_display import display_image_safe

import numpy as np
import cv2 as cv


# Change this value based on the Kart. VALUES ARE 1 OR 2!!!!
CURRENT_KART = 1
# ----------------------------------------------------------

# Store a set of previous values to return if zebra/hor line detected
m_left_old = 150
m_right_old = 150
m_left_2m_old = 150
m_right_2m_old = 150
# -------------------------------------------------------------------
# When a horizontal line at this dist, start returnng old values
HORIZONTAL_DIST_CUTOFF = 200


class LaneResult(TypedDict):
    dist: Optional[float]
    type: Optional[str]


class DetectedLaneResult(TypedDict):
    left: LaneResult
    right: LaneResult
    far_left: LaneResult
    far_right: LaneResult
    horizontal: LaneResult


def _get_lane_dist_from_image(image: cv.Mat) -> tuple[DetectedLaneResult, Optional[int]]:
    """
    Wrapper Method. Detects left/right lanes using a frame from Main Camera and return distance to them in a standard format.
    :param image: Frame from the main car camera -> cv.Mat
    :return: standard message -> DetectedLaneResult
    """
    global m_left_old
    global m_right_old
    global m_left_2m_old
    global m_right_2m_old

    img_prepped = _img_prep(image)
    bev = _to_birds_eye_view(img_prepped)
    thresh = _threshold_lanes(bev)

    # display_image_safe(bev, "Birds Eye View")
    # display_image_safe(thresh, "Thresh Main")

    # Return old values if horizontal line/ zebra detected
    dist_hor, type_hor = detect_horizontal_line(bev)
    # if type_hor is not None and dist_hor <= HORIZONTAL_DIST_CUTOFF:
    #     return make_ros_message(m_left_old, m_right_old, m_left_2m_old, m_right_2m_old, dist_hor, type_hor)

    m_left, m_right, m_left_2m, m_right_2m, right_front_track_x = _track_lanes(
        thresh)

    # Round values to closest 10 to get a more stable output. Return None if a value is less than 0
    if m_left is not None and m_left < 0:
        m_left = None
    if m_right is not None and m_right < 0:
        m_right = None
    if m_left_2m is not None and m_left_2m < 0:
        m_left_2m = None
    if m_right_2m is not None and m_right_2m < 0:
        m_right_2m = None

    # Round values to closest 10 to get a more stable output. Return None if a value is less than 0
    if m_left is not None and m_left < 0:
        m_left = None
    if m_right is not None and m_right < 0:
        m_right = None
    if m_left_2m is not None and m_left_2m < 0:
        m_left_2m = None
    if m_right_2m is not None and m_right_2m < 0:
        m_right_2m = None

    # Round values to closest 10 to get a more stable output. Return 0 if a value is less than 0
    if m_left is not None:
        m_left = round(m_left / 10) * 10
    if m_right is not None:
        m_right = round(m_right / 10) * 10
    if m_left_2m is not None:
        m_left_2m = round(m_left_2m / 10) * 10
    if m_right_2m is not None:
        m_right_2m = round(m_right_2m / 10) * 10

    # Save the latest values
    m_left_old = m_left
    m_right_old = m_right
    m_left_2m_old = m_left_2m
    m_right_2m_old = m_right_2m

    return make_ros_message(m_left, m_right, m_left_2m, m_right_2m, dist_hor, type_hor), right_front_track_x


def _get_single_lane_distance(image: cv.Mat, is_left_lane: bool = True) -> Optional[float]:
    """
    Wrapper Method. Detects left/right single lane using a frame from Main Camera and return distance to it
    :param image: Frame from the side car camera -> cv.Mat
    :return: distance in meters -> float
    """
    img_prepped = _img_prep(image)
    if is_left_lane is True:
        bev = _left_side_cam_birds_eye_view(img_prepped, CURRENT_KART)
    else:
        bev = _right_side_cam_birds_eye_view(img_prepped, CURRENT_KART)

    thresh = _threshold_lanes(bev)

    # display_image_safe(
    #     bev, f"Birds Eye View Side {'Left' if is_left_lane else 'Right'}"
    # )

    m_dist = _track_single_lane(thresh, is_left_lane)

    # Round values to closest 10 to get a more stable output. Return None if a value is less than 0
    if m_dist is not None and m_dist < 0:
        m_dist = None

    if m_dist is not None:
        m_dist = round(m_dist / 10) * 10

    return m_dist


def make_ros_message(
    m_left: Optional[float], m_right: Optional[float],
    m_left_2m: Optional[float], m_right_2m: Optional[float],
    dist_hor: Optional[float], type_hor: Optional[str],
    type_left: Optional[str] = None, type_right: Optional[str] = None
) -> DetectedLaneResult:
    message: DetectedLaneResult = {
        'left': {
            'dist': m_left,
            'type': type_left,
        },
        'right': {
            'dist': m_right,
            'type': type_right,
        },
        'far_left': {
            'dist': m_left_2m,
            'type': None,
        },
        'far_right': {
            'dist': m_right_2m,
            'type': None,
        },
        'horizontal': {
            'dist': dist_hor,
            'type': type_hor,
        },
    }

    return message


def _load_path_to_img_obj(image_path: str) -> cv.Mat:
    """
    Creates an image obj from input
    :param image_path:
    :return: Mat obj
    """
    image_obj = cv.imread(image_path)
    # coord 848x480
    image_obj = cv.resize(image_obj, (848, 480))
    return image_obj


def _img_prep(image: cv.Mat) -> cv.UMat:
    """
    Converts the input into a gray scaled & blurred image for later Comp Vision
    :param image: Image from car camera
    :return: Gray scaled and blurred image -> UMat obj
    """
    # coord 848x480
    image = cv.resize(image, (848, 480))
    gray = cv.cvtColor(image, cv.COLOR_RGB2GRAY)

    # use gaussian blur to smoothen image and reduce blur
    # Minimize thresholding noise of img
    blur = cv.GaussianBlur(gray, (5, 5), 0)
    return blur


def _threshold_lanes(image: cv.UMat) -> cv.Mat:
    """
    Find white edges in image
    :param image: from cart camera. Must be warped to Birds Eye View
    :return: binary image of edges -> cv.Mat
    """
    # Isolate white from the image to highlight the white lines
    _, white_mask = cv.threshold(image, 190, 255, cv.THRESH_BINARY)
    white_isolated = cv.bitwise_and(image, image, mask=white_mask)
    # Find edges using canny thresholding. Convert the image to a binary array
    # 1 if an edge is present, 0 if no edge is present
    canny = cv.Canny(white_isolated, 50, 50)
    return canny


def _track_single_lane(binary_image: cv.Mat, is_left_lane: bool = True) -> Optional[float]:
    '''
    Tracks single white lane using the Sliding Window Technqiue. Finds distance to right/left lane by fitting a polynom to
    tracked lanes
    :param is_left_lane: specifies if right or left side of frame should be used. By default uses left side of frame,
    because there are only left turn on the track
    :param binary_image: thresholded image in binary format
    :return: distance to left/right lane meters -> float
    '''
    # get a histogram of the bottom half of the image
    histogram = np.sum(binary_image[int(binary_image.shape[0]/2):, :], axis=0)

    # make output image to visualize result
    out_img = np.dstack((binary_image, binary_image, binary_image))*255

    # func to get all peaks in histogram

    def find_peaks(inp_hist):
        peaks = []
        for i in range(1, len(inp_hist) - 1):
            if inp_hist[i] > inp_hist[i - 1] and inp_hist[i] > inp_hist[i + 1]:
                peaks.append(i)
        return peaks

    # Find all peaks
    peaks = find_peaks(histogram)

    # Find peak of the left and right halves of the histogram -> starting points for left/right lane
    midpoint = int(histogram.shape[0]/2)
    if is_left_lane is True:
        if len(peaks) == 0:
            x_base = np.argmax(histogram[:midpoint])  # left side of frame
        else:
            # Find the left most peak
            left_peaks = [p for p in peaks if p < midpoint]
            if len(left_peaks) > 0:
                x_base = min(left_peaks, key=lambda x: x)
            else:
                x_base = np.argmax(histogram[:midpoint])
    else:
        if len(peaks) == 0:
            # right side of frame
            x_base = np.argmax(histogram[midpoint:]) + midpoint
        else:
            # Find the rightmost peak
            right_peaks = [p for p in peaks if p >= midpoint]
            if len(right_peaks) > 0:
                x_base = min(right_peaks, key=lambda x: len(histogram) - 1 - x)
            else:
                x_base = np.argmax(histogram[midpoint:]) + midpoint

    # num of sliding windows per lane to track midpoint concentration of pixels
    num_windows = 9
    window_height = int(binary_image.shape[0]/num_windows)
    # width of the windows +/- margin
    margin = 100
    # min number of pixels found to recenter window
    minpix = 50

    # get the x and y positions of all nonzero pixels in the image
    nonzero = binary_image.nonzero()
    nonzeroy = np.array(nonzero[0])
    nonzerox = np.array(nonzero[1])

    # current position to be updated for each window
    x_current = x_base

    # left/right lane pixel indices
    lane_inds = []

    # step through the windows one by one
    for window in range(num_windows):
        # identify window boundaries in x and y (and right and left)
        win_y_low = binary_image.shape[0] - (window+1)*window_height
        win_y_high = binary_image.shape[0] - window*window_height
        win_x_low = x_current - margin
        win_x_high = x_current + margin

        # Draw the windows on the image
        cv.rectangle(out_img, (win_x_low, win_y_low),
                     (win_x_high, win_y_high), (0, 255, 0), 2)

        # Identify the nonzero pixels in x and y within the window
        good_inds = ((nonzeroy >= win_y_low) & (nonzeroy < win_y_high) & (
            nonzerox >= win_x_low) & (nonzerox < win_x_high)).nonzero()[0]

        # Append these indices to the lists
        lane_inds.append(good_inds)

        # If you found > minpix pixels, recenter next window on their mean position
        if len(good_inds) > minpix:
            x_current = int(np.mean(nonzerox[good_inds]))

    # concatenate the arrays of indices
    lane_inds = np.concatenate(lane_inds)

    # Extract left and right line pixel positions
    # Coordinates from bottom of frame to top
    x = nonzerox[lane_inds]
    y = nonzeroy[lane_inds]

    # First get distance using a less reliable method -> look at X coords of found thresholds. Not accurate, cannot predict lanes
    if is_left_lane is True:
        m_dist = _get_left_distance_to_lanes(x,)
    else:
        m_dist = _get_right_distance_to_lane(x,)

    # ----------------------- Get distance by fitting a polynomial -> more reliable ------------------------
    # Generate y values for plotting (0-799)
    ploty = np.linspace(0, binary_image.shape[0]-1, binary_image.shape[0])
    # Create an image to draw the lines on
    warp_zero = np.zeros_like(binary_image).astype(np.uint8)

    coordinates_list = []
    intercept = None
    fitx = []

    if (len(x) != 0) | (len(y) != 0):
        # fit a second order polynomial (ax^2 +bx +c) to lane. Can later be used to find an intersection with bottom of frame.
        # Allows to have a general estimation of a line, even if parts of it are not found by thresholding
        fit = np.polyfit(y, x, 2)
        fitx = fit[0]*ploty**2 + fit[1]*ploty + fit[2]
        # Recast the x and y points into usable format for cv.fillPoly()
        pts = np.array([np.transpose(np.vstack([fitx, ploty]))])
        coordinates_list.append(pts)

    # Check if intercepts were found here to minimize the number of times a function is ran
    if len(fitx) != 0:
        if is_left_lane is True:
            # Get interception of polynom with y coord of the edge point of bumper
            intercept = fitx[len(fitx) - 200]
            m_dist = _get_left_distance_to_lanes_coordinate(intercept)
        else:
            # Get interception of polynom with y coord of the edge point of bumper
            intercept = fitx[len(fitx) - 200]
            m_dist = _get_right_distance_to_lanes_coordinate(intercept)

    return m_dist


def detect_horizontal_line(
    bev_image: cv.UMat, leftfit: Optional[np.ndarray] = None, rightfit: Optional[np.ndarray] = None
) -> tuple[Optional[float], Optional[str]]:
    """
    Detects horizontal lines ahead and classifies them in a standard format. Uses amount of white in the middle
    of the road.
    :param rightfit: array with X values of the curve fit to the right detected lane -> ndarray
    :param leftfit: array with X values of the curve fit to the left detected lane -> ndarray
    :param bev_image: Main car camera. Must be warped to Birds Eye View -> cv.UMat
    :return: type of horizontal line and distance to it in px -> tuple[dist, type]
    """
    ZEBRA_THRESHOLD = 20  # percentage of white pixels when a zebra pad is in frame
    LINE_THRESHOLD = 10  # percentage of white pixels when a horizontal lane is in frame

    def get_x_value(y_val_top_left: int, y_val_bottom_right: int) -> tuple[int | None, int | None]:
        if leftfit is not None:
            x1 = leftfit[y_val_top_left]
        else:
            x1 = 0

        if rightfit is not None:
            x2 = rightfit[y_val_bottom_right]
        else:
            x2 = bev_image.shape[1]
        return x1, x2

    # Check middle of road using a sliding horizontal rectangle with h=100px. Go bottom->top
    window_height = 100
    i = 0  # number of the window from the bottom of the frame
    dist = None
    lane_type = None
    percentages = []
    for a in range((bev_image.shape[0] // window_height) - 1):
        y2 = bev_image.shape[0] - (window_height * i)  # bottom of ROI
        y1 = y2 - window_height  # top of ROI
        x1, x2 = get_x_value(y1, y2)
        roi = bev_image[y1:y2, x1:x2]  # get ROI

        # Isolate white from the image to find white lines
        _, white_mask = cv.threshold(roi, 190, 255, cv.THRESH_BINARY)
        # Get ratio of white to non-white pixels
        # get total number of pixels in the mask
        total_pixels = white_mask.shape[1] * white_mask.shape[0]
        white_pixels = cv.countNonZero(white_mask)  # get num of white pixels
        white_percentage = (white_pixels / total_pixels) * \
            100  # get percentage of white pixels
        i = i + 1
        if white_percentage >= ZEBRA_THRESHOLD:
            dist = (i * window_height) - window_height
            lane_type = 'zebra'
            return dist, lane_type
        if white_percentage >= LINE_THRESHOLD:
            dist = (i * window_height) - window_height
            lane_type = 'horizontal'
            return dist, lane_type

    return dist, lane_type


def _detect_grass_percentage(image: cv.Mat, is_left_camera: bool = True) -> float:
    '''
    Return a percentage of grass (green) found in the side camera frame
    :param image: Frame from side camera -> cv.Mat
    :param is_left_camera: Indicate whether image from left/right camera -> bool
    :return: percentage of green in a frame -> float
    '''
    image = cv.resize(image, (848, 480))
    if is_left_camera:
        image_bev = _left_side_cam_birds_eye_view(image, CURRENT_KART)
    else:
        image_bev = _right_side_cam_birds_eye_view(image, CURRENT_KART)

    # Converts BGR to HSV
    hsv = cv.cvtColor(image_bev, cv.COLOR_BGR2HSV)

    # define the range of green
    lower_green = np.array([40, 40, 40])
    upper_green = np.array([70, 255, 255])

    # Threshold the HSV image to filter out green
    mask = cv.inRange(hsv, lower_green, upper_green)

    # Get ratio of green to non-green pixels
    # get total number of pixels in the mask
    total_pixels = mask.shape[1] * mask.shape[0]
    green_pixels = cv.countNonZero(mask)  # get num of green pixels
    green_percentage = (green_pixels / total_pixels) * \
        100  # get percentage of green pixels

    return green_percentage
