import cv2 as cv
import numpy as np

# Values for BEV transformation. Values get updated from the bev_values.json
BEV_MATRIX_FRONT = {
    'src': np.float32([[432, 249], [627, 249], [432, 267], [576, 295]]),
    'dst': np.float32([[0, 0], [640 - 1, 0], [0, 800 - 1], [640 - 1, 800 - 1]]),
}


def _to_birds_eye_view(image: cv.UMat) -> cv.UMat:
    """
    Warps an UMat image object into a birds eye view with 3m ahead and 1.5m on each side.
    :param image: Image from car camera. Size 848x480 pixels!
    :return: warped image of cv.UMat
    """
    global BEV_MATRIX_FRONT
    # width, height of output image
    image_size = (640, 800)
    # Input box to be transformed. Input IMAGE SIZE MUST BE 848x480 pix!!!
    src = BEV_MATRIX_FRONT['src']
    # Each point from src get shifted to following coords
    dst = BEV_MATRIX_FRONT['dst']

    M = cv.getPerspectiveTransform(src, dst)
    warped_img = cv.warpPerspective(image, M, image_size)
    return warped_img


def _track_lanes(binary_image: cv.Mat) -> tuple[None | float, None | float, None | float, None | float]:
    '''
    Tracks white lanes using the Sliding Window Technqiue. Finds distances to right/left lane by fitting a polynom to
    tracked lanes
    :param binary_image: thresholded image in binary format
    :return: distance to left/right lane from car and dist to left/right 2m ahead -> float
    '''
    # get a histogram of the bottom half of the image
    histogram = np.sum(binary_image[int(binary_image.shape[0]/2):, :], axis=0)

    # make output image to visualize result
    out_img = np.dstack((binary_image, binary_image, binary_image))*255

    # func to get all peaks in histogram
    def find_peaks(inp_hist):
        peaks = []
        for i in range(1, len(inp_hist) - 1):
            if inp_hist[i] > inp_hist[i - 1] and inp_hist[i] > inp_hist[i + 1]:
                peaks.append(i)
        return peaks

    # Find all peaks
    peaks = find_peaks(histogram)

    # If no peaks are found, handle this case (optional)
    if len(peaks) == 0:
        # Find peak of the left and right halves of the histogram -> starting points for left/right lane
        midpoint = int(histogram.shape[0]/2)
        leftx_base = np.argmax(histogram[:midpoint])
        rightx_base = np.argmax(histogram[midpoint:]) + midpoint
    else:
        # Calculate midpoint
        midpoint = int(histogram.shape[0] / 2)

        # Find the leftmost peak
        left_peaks = [p for p in peaks if p < midpoint]
        if len(left_peaks) > 0:
            leftx_base = min(left_peaks, key=lambda x: x)
        else:
            leftx_base = np.argmax(histogram[:midpoint])

        # Find the rightmost peak
        right_peaks = [p for p in peaks if p >= midpoint]
        if len(right_peaks) > 0:
            rightx_base = min(right_peaks, key=lambda x: len(histogram) - 1 - x)
        else:
            rightx_base = np.argmax(histogram[midpoint:]) + midpoint

    # Find peak of the left and right halves of the histogram -> starting points for left/right lane
    # midpoint = int(histogram.shape[0]/2)
    # leftx_base = np.argmax(histogram[:midpoint])
    # rightx_base = np.argmax(histogram[midpoint:]) + midpoint

    # num of sliding windows per lane to track midpoint concentration of pixels
    num_windows = 9
    window_height = int(binary_image.shape[0]/num_windows)
    # width of the windows +/- margin
    margin = 100
    # min number of pixels found to recenter window
    minpix = 50

    # get the x and y positions of all nonzero pixels in the image
    nonzero = binary_image.nonzero()
    nonzeroy = np.array(nonzero[0])
    nonzerox = np.array(nonzero[1])

    # current positions to be updated for each window
    leftx_current = leftx_base
    rightx_current = rightx_base

    # left/right lane pixel indices
    left_lane_inds = []
    right_lane_inds = []

    # step through the windows one by one
    for window in range(num_windows):
        # identify window boundaries in x and y (and right and left)
        win_y_low = binary_image.shape[0] - (window+1)*window_height
        win_y_high = binary_image.shape[0] - window*window_height
        win_xleft_low = leftx_current - margin
        win_xleft_high = leftx_current + margin
        win_xright_low = rightx_current - margin
        win_xright_high = rightx_current + margin

        # Draw the windows on the image
        cv.rectangle(out_img, (win_xleft_low, win_y_low),
                     (win_xleft_high, win_y_high), (0, 255, 0), 2)
        cv.rectangle(out_img, (win_xright_low, win_y_low),
                     (win_xright_high, win_y_high), (0, 255, 0), 2)

        # Identify the nonzero pixels in x and y within the window
        good_left_inds = ((nonzeroy >= win_y_low) & (nonzeroy < win_y_high) & (
            nonzerox >= win_xleft_low) & (nonzerox < win_xleft_high)).nonzero()[0]
        good_right_inds = ((nonzeroy >= win_y_low) & (nonzeroy < win_y_high) & (
            nonzerox >= win_xright_low) & (nonzerox < win_xright_high)).nonzero()[0]

        # Append these indices to the lists
        left_lane_inds.append(good_left_inds)
        right_lane_inds.append(good_right_inds)

        # If you found > minpix pixels, recenter next window on their mean position
        if len(good_left_inds) > minpix:
            leftx_current = int(np.mean(nonzerox[good_left_inds]))
        if len(good_right_inds) > minpix:
            rightx_current = int(np.mean(nonzerox[good_right_inds]))

    # concatenate the arrays of indices
    left_lane_inds = np.concatenate(left_lane_inds)
    right_lane_inds = np.concatenate(right_lane_inds)

    # Extract left and right line pixel positions
    # Coordinates from bottom of frame to top
    leftx = nonzerox[left_lane_inds]
    lefty = nonzeroy[left_lane_inds]
    rightx = nonzerox[right_lane_inds]
    righty = nonzeroy[right_lane_inds]

    # First get distance using a less reliable method -> look at X coords of found thresholds. Not accurate, cannot predict lanes
    m_left, m_right = _get_distance_to_lanes(leftx, rightx, binary_image)

    # ----------------------- Get distance by fitting a polynomial -> more reliable ------------------------
    # Generate y values for plotting (0-799)
    ploty = np.linspace(0, binary_image.shape[0]-1, binary_image.shape[0])
    # Create an image to draw the lines on
    warp_zero = np.zeros_like(binary_image).astype(np.uint8)

    coordinates_list = []
    left_intercept_2m = None
    right_intercept_2m = None
    left_intercept = None
    right_intercept = None
    m_left_2m = None
    m_right_2m = None
    
    right_front_track_x = None

    if (len(leftx) != 0) | (len(lefty) != 0):
        # Case where left lane is found
        # fit a second order polynomial (ax^2 +bx +c) to lane. Can later be used to find an intersection with bottom of frame.
        # Allows to have a general estimation of a line, even if parts of it are not found by thresholding
        left_fit = np.polyfit(lefty, leftx, 2)
        left_fitx = left_fit[0]*ploty**2 + left_fit[1]*ploty + left_fit[2]
        # Recast the x and y points into usable format for cv.fillPoly()
        pts_left = np.array([np.transpose(np.vstack([left_fitx, ploty]))])
        coordinates_list.append(pts_left)
        # Get interception of polynom with bottom of frame
        left_intercept = left_fitx[len(left_fitx) - 200]
        # Get the graph values 2m ahead of the car at y level = 270. 800 - (800*2/3) = 270
        left_intercept_2m = left_fitx[270]        


    if (len(rightx) != 0) | (len(righty) != 0):
        # Case where right lane is found
        # fit a second order polynomial (ax^2 +bx +c) to lane. Can later be used to find an intersection with bottom of frame.
        # Allows to have a general estimation of a line, even if parts of it are not found by thresholding
        right_fit = np.polyfit(righty, rightx, 2)
        right_fitx = right_fit[0]*ploty**2 + right_fit[1]*ploty + right_fit[2]
        # Recast the x and y points into usable format for cv.fillPoly()
        pts_right = np.array(
            [np.flipud(np.transpose(np.vstack([right_fitx, ploty])))])
        coordinates_list.append(pts_right)
        # Get interception of polynom with bottom of frame
        right_intercept = right_fitx[len(right_fitx) - 200]
        # Get the graph values 2m ahead of the car at y level = 270. 800 - (800*2/3) = 270
        right_intercept_2m = right_fitx[270]
        right_front_track_x = right_fitx[400]


    # Check if intercepts were found here to minimize the number of times a function is ran
    if (left_intercept is not None) and (right_intercept is not None):
        m_left, m_right = _get_distance_to_lanes_coordinate(
            left_intercept, right_intercept, binary_image)
        m_left_2m, m_right_2m = _get_distance_to_lanes_coordinate(
            left_intercept_2m, right_intercept_2m, binary_image)

    elif (left_intercept is None):
        _, m_right = _get_distance_to_lanes_coordinate(
            left_intercept, right_intercept, binary_image)
        _, m_right_2m = _get_distance_to_lanes_coordinate(
            left_intercept_2m, right_intercept_2m, binary_image)
    elif (right_intercept is None):
        m_left, _ = _get_distance_to_lanes_coordinate(
            left_intercept, right_intercept, binary_image)
        m_left_2m, _ = _get_distance_to_lanes_coordinate(
            left_intercept_2m, right_intercept_2m, binary_image)

    # return m_left, m_right
    return m_left, m_right, m_left_2m, m_right_2m, right_front_track_x


def _get_distance_to_lanes(left: np.array, right: np.array, binary_image: cv.Mat) -> tuple[None | float, None | float]:
    '''
    Calculates the distance from the middle of the frame to the left/right lane.
    Caluclations done using the bottom row of the pixels
    :param binary_image:
    :param left: left line pixel positions -> np.array
    :param right: right line pixel positions -> np.array
    :return: left, right distance to lanes in meters -> tuple[float, float]
    '''
    dist_left = None
    dist_right = None
    # get pixel coordinate of the middle of the frame (mid of the car)
    try:
        frame_middle = binary_image.shape[1] // 2
    except IndexError:
        return dist_left, dist_right

    # get distance from middle of the frame to left lane
    try:
        dist_left = frame_middle - left[0]
    except IndexError:
        # if no image found, return None as distance
        pass

    # get distance from middle of the frame to left lane
    try:
        dist_right = right[0] - frame_middle
    except IndexError:
        # if no image found, return None as distance
        pass

    m_left, m_right = _calc_distance_meters(dist_left, dist_right)
    return m_left, m_right


def _get_distance_to_lanes_coordinate(left: float | None, right: float | None, binary_image: cv.Mat) -> tuple[None | float, None | float]:
    '''
    Calculates the distance from the middle of the frame to the left/right lane.
    Caluclations done using an input left/right x coordinate of a polynom intercept with bottom of frame
    :param binary_image:
    :param left: pixel position of left intercept-> float
    :param right: pixel position of right intercept-> float
    :return: left, right distance to lanes in meters -> tuple[float, float]
    '''
    dist_left = None
    dist_right = None
    # get pixel coordinate of the middle of the frame (mid of the car)
    try:
        frame_middle = binary_image.shape[1] // 2
    except IndexError:
        return dist_left, dist_right

    try:
        dist_left = frame_middle - left
    except:
        pass

    try:
        dist_right = right - frame_middle
    except:
        pass

    m_left, m_right = _calc_distance_meters(dist_left, dist_right)
    return m_left, m_right


# Removed calculation to meters Temporary
def _calc_distance_meters(dist_left: float | None, dist_right: float | None) -> tuple[None | float, None | float]:
    '''
    Converts distance in pixels to distance in meters for main camera
    :param dist_left: distance to left lane in px -> float
    :param dist_right: distance to right lane in px -> float
    :return: real world distance in meters to left/right lane -> tuple[float, float]
    '''
    # after calculation this is the scale factor that can be used to transform px to m
    # new_distance_left, new_distance_right = None, None
    # if dist_left is not None and dist_right is not None:
    #     # The road width along the track is always 3m
    #     scale_factor = 3 / (dist_right + dist_left)  # 3m is width of track
    # else:
    #     # Back up estimated scale factor
    #     scale_factor = 0.0071
    #
    # if dist_left is not None:
    #     new_distance_left = dist_left * scale_factor
    #
    # if dist_right is not None:
    #     new_distance_right = dist_right * scale_factor
    #
    # return new_distance_left, new_distance_right
    return dist_left, dist_right
