import cv2 as cv
import numpy as np

# Birds Eye View Transformations Values for Karts 1 & 2
perspective_transforms_left = {
    1: {
        'src': np.float32([[610, 110], [848, 220], [100, 70], [848, 540]]),
        'dst': np.float32([[0, 0], [640 - 1, 0], [0, 800 - 1], [640 - 1, 800 - 1]]),
    },
    2: {
        'src': np.float32([[710, 110], [848, 180], [100, 120], [848, 480]]),
        'dst': np.float32([[0, 0], [640 - 1, 0], [0, 800 - 1], [640 - 1, 800 - 1]]),
    }
}


def _left_side_cam_birds_eye_view(image: cv.UMat, kart: int = 1) -> cv.UMat:
    """
    Warps an UMat image object from left side camera into a birds eye view.
    :param image: Image from car left side camera. Size 848x480 pixels!
    :param kart: kart number 1 or 2. Default kart is 1
    :return: warped image of cv.UMat
    """
    global perspective_transforms_left
    # width, height of output image
    image_size = (640, 800)

    # Initial road pane defined by 4 points
    src = perspective_transforms_left[kart]['src']
    # Each point from src get shifted to following coords
    dst = perspective_transforms_left[kart]['dst']

    M = cv.getPerspectiveTransform(src, dst)
    warped_img = cv.warpPerspective(image, M, image_size)
    return warped_img


# Removed meter calculation. Return pixel values now
def _calc_distance_meters_left(dist_left: float | None) -> float | None:
    """
    Converts distance in pixels to distance in meters for left camera
    :param dist_left: distance to right lane in px -> float
    :return: real world distance in meters to left lane -> float
    """
    # # after calculation this is the scale factor that can be used to transform px to m
    # new_distance_left = None
    # scale_factor = 0.004  # based on width of white lane
    #
    # if dist_left is not None:
    #     new_distance_left = dist_left * scale_factor
    #     # Add the distance from corner of front bumper to middle of the car
    #     # 0.49m is half the distance of bumper
    #     new_distance_left = new_distance_left + 0.49
    # return new_distance_left
    if dist_left is not None:
        scale_factor = 0.004  # based on width of white lane
        bumper_size = 0.49  # 0.49m is half the distance of bumper
        dist_left = dist_left + (bumper_size//scale_factor)
    return dist_left


def _get_left_distance_to_lanes(left: np.array) -> None | float:
    '''
    Calculates the distance from the middle of the car to the left lane.
    Caluclations done using the bottom row of the pixels
    :param binary_image:
    :param left: left line pixel positions of the Threshold -> np.array
    :return: left, distance to lanes in meters from middle of the car -> float
    '''
    dist_left = None
    # get distance from middle of the frame to left lane
    try:
        dist_left = 245 - left[0]
    except IndexError:
        # if no image found, return None as distance
        return None

    m_left = _calc_distance_meters_left(dist_left)
    return m_left


def _get_left_distance_to_lanes_coordinate(left: float | None) -> None | float:
    '''
    Calculates the distance from the middle of the car to the left lane.
    Caluclations done using an input left x coordinate of a polynom intercept with y value of bumper
    :param left: pixel position of left intercept-> float
    :return: left, distance to lanes in meters -> float
    '''
    dist_left = None

    try:
        dist_left = 245 - left
    except:
        return None

    m_left = _calc_distance_meters_left(dist_left)
    return m_left
