import cv2 as cv
import numpy as np

# Birds Eye View Transformations Values for Karts 1 & 2
perspective_transforms_right = {
    1: {
        'src': np.float32([[0, 250], [110, 190], [0, 480], [848, 105]]),
        'dst': np.float32([[0, 0], [640 - 1, 0], [0, 800 - 1], [640 - 1, 800 - 1]]),
    },
    2: {
        'src': np.float32([[0, 220], [115, 160], [0, 480], [848, 83]]),
        'dst': np.float32([[0, 0], [640 - 1, 0], [0, 800 - 1], [640 - 1, 800 - 1]]),
    }
}


def _right_side_cam_birds_eye_view(image: cv.UMat, kart: int = 1) -> cv.UMat:
    """
    Warps an UMat image object from right side camera into a birds eye view.
    :param image: Image from car side camera. Size 848x480 pixels!
    :param kart: kart number 1 or 2. Default kart is 1
    :return: warped image of cv.UMat
    """
    global perspective_transforms_right
    # width, height of output image
    image_size = (640, 800)

    # Initial road pane defined by 4 points
    src = perspective_transforms_right[kart]['src']
    # # Each point from src get shifted to following coords
    dst = perspective_transforms_right[kart]['dst']

    M = cv.getPerspectiveTransform(src, dst)
    warped_img = cv.warpPerspective(image, M, image_size)
    return warped_img


# Removed meter calculation. Return pixel values now
def _calc_distance_meters_right(dist_right: float | None) -> float | None:
    """
    Converts distance in pixels to distance in meters for right camera
    :param dist_right: distance to right lane in px -> float
    :return: real world distance in meters to right lane -> float
    """
    # after calculation this is the scale factor that can be used to transform px to m
    # new_distance_right = None
    # scale_factor = 0.0036  # based on width of white lane
    #
    # if dist_right is not None:
    #     new_distance_right = dist_right * scale_factor
    #     # Add the distance from corner of front bumper to middle of the car
    #     new_distance_right = new_distance_right + \
    #         0.49  # 0.49m is half the distance of bumper
    # return new_distance_right
    if dist_right is not None:
        scale_factor = 0.0036  # based on width of white lane
        bumper_size = 0.49  # 0.49m is half the distance of bumper
        dist_right = dist_right + (bumper_size//scale_factor)
    return dist_right


def _get_right_distance_to_lane(right: np.array) -> None | float:
    '''
   Calculates the distance from the middle of the car to the right lane.
   Caluclations done using the bottom row of the pixels
   :param right: right line pixel positions of the Threshold -> np.array
   :return: right distance to lanes in meters from middle of the car -> float
   '''
    dist_right = None
    # get distance from middle of the frame to left lane
    try:
        dist_right = right[0] - 240
    except IndexError:
        # if no image found, return None as distance
        pass

    m_right = _calc_distance_meters_right(dist_right)
    return m_right


def _get_right_distance_to_lanes_coordinate(right: float | None) -> None | float:
    '''
    Calculates the distance from the middle of the car to the right lane.
    Caluclations done using an input left/right x coordinate of a polynom intercept with y value of bumper
    :param right: pixel position of right intercept-> float
    :return: right distance to lanes in meters -> float
    '''

    dist_right = None
    try:
        dist_right = right - 240
    except:
        pass

    m_right = _calc_distance_meters_right(dist_right)
    return m_right
