from rclpy.node import Node
from custom_interfaces.msg import PriorityTwist, Lanes
from std_msgs.msg import Float32
from geometry_msgs.msg import Twist
from sensor_msgs.msg import LaserScan

from core.lane_tracking.utils import DetectedLaneResult
from ..utils import LidarProcessor
from .utils import OvertackingPlanning, LaneSwitcher
from .overtaking_methods import OvertakingMethods

# from .overtaking_methods import OvertakingMethods
import rclpy


class OvertakingNode(Node):
    # the amount of values the lidar must return
    # (in practice the Lidar does not always return 1080)
    VALID_LIDAR_VALUES_NUM = 1080

    def __init__(self):
        super().__init__('overtaking_node')
        self.get_logger().info('OvertakingNode started')

        # Subscribe to the /scan topic
        self.create_subscription(
            LaserScan,
            '/scan',
            self._scan_callback,
            10
        )

        self.create_subscription(
            Float32,
            '/lane_x_position',
            self._lane_x_position_callback,
            10
        )

        self.create_subscription(
            Lanes,
            '/lane_detections',
            self._lane_detections_callback,
            10
        )
        self._detected_lane: DetectedLaneResult = {}

        # Publish to the /priority_cmd_vel topic
        self._cmd_vel_publisher = self.create_publisher(
            PriorityTwist,
            '/priority_cmd_vel',
            10
        )
        self._priority_twist_message: PriorityTwist = PriorityTwist()

        self. _lane_x_position = 0

        self._overtaking_methods = OvertakingMethods()

        # Used for the "alternative way of overtaking"
        self._processor = LidarProcessor()
        self._planner = OvertackingPlanning(
            logger=self.get_logger(),
            get_angle_distance=self._processor.get_angle_distance
        )
        self._lane_switcher = LaneSwitcher(
            logger=self.get_logger()
        )

    def _scan_callback(self, msg: LaserScan) -> None:
        # Ignore the message if the number of values is not correct
        if (len(msg.ranges) != self.VALID_LIDAR_VALUES_NUM):
            return

        # ===================================================================================================================
        # Alternative way of overtaking ====================================================================================
        # ===================================================================================================================
        # Process the scan
        objects = self._processor.process_scan(msg)
        tracked_objects = self._processor.track_detected_objects(objects)

        # Plan the overtaking maneuver
        twist_message = self._planner.plan(
            tracked_objects, self._detected_lane
        )

        if twist_message is not None:
            self.publish_velocity(twist_message)
        # ===================================================================================================================
        # ===================================================================================================================
        # ===================================================================================================================

        # twist_message = self._overtaking_methods.perform_overtaking(  # , right_count, object_location, distance_to_object
        #     msg,
        #     self._lane_x_position
        # )

        # if twist_message is not None:
        #     self.get_logger().info(
        #         f"""
        #             linear: {twist_message.linear.x}, angular: {twist_message.angular.z}
        #             testing variable: {self._overtaking_methods.testing_variable}
        #             is_overtake_done: {self._overtaking_methods.is_overtake_done}
        #         """
        #     )

        #     self.publish_velocity(twist_message)

    def _lane_x_position_callback(self, msg: Float32) -> None:
        self._lane_x_position = msg.data

    def _lane_detections_callback(self, msg: Lanes) -> None:
        self._detected_lane: DetectedLaneResult = {
            'left': {
                'dist': msg.left.dist,
                'type': msg.left.type
            },
            'right': {
                'dist': msg.right.dist,
                'type': msg.right.type
            },
            'far_left': {
                'dist': msg.far_left.dist,
                'type': msg.far_left.type
            },
            'far_right': {
                'dist': msg.far_right.dist,
                'type': msg.far_right.type

            },
            'horizontal': {
                'dist': msg.horizontal.dist,
                'type': msg.horizontal.type
            }
        }

    def publish_velocity(self, twist_message: Twist, priority: int = 3) -> None:
        self._priority_twist_message.priority = priority
        self._priority_twist_message.twist = twist_message

        self._cmd_vel_publisher.publish(self._priority_twist_message)


def main(args=None):
    rclpy.init(args=args)
    node = OvertakingNode()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
