from gc import is_finalized
import re
from typing import Optional
from simple_pid import PID
import numpy as np

from geometry_msgs.msg import Twist
from sensor_msgs.msg import LaserScan

from ..utils import Movement, ObjectDetection, DetectionLocation


class OvertakingMethods(object):
    MAX_RIGHT_COUNT = 40
    LANE_X_THRESHOLD = 100
    MAX_DIST_TO_LANE = 800

    # Lidar uses this key (206 degrees) to stay OVERTAKE_DISTANCE away from
    # object with that azimuth
    WAYPOINT_ANGLE = 715
    # Kart will stay away this dist from the other car (in m)
    OVERTAKE_DISTANCE = 2.5
    # Lidar uses this key to know when its next to a car to engage PID controller 
    NEXT_TO_CAR_ANGLE = 910
    # Lidar uses this key to check if it overtook the car
    OVERTAKE_CUTOFF_ANGLE = 835

    MEMORY_BUFFER_LEN = 25 # amount of past readings from lidar to be saved

    def __init__(self):
        self._movement = Movement()
        self._object_detection = ObjectDetection()
        self._right_count = 0
        self.enough_rotations = False
        self.is_car_detected = False
        self.is_in_left_lane = False
        self.is_next_to_car = False
        self.is_overtake_done = False
        self.is_finished = False
        self.is_lanetrack_ready = False
        self.initial_smallest_key = None

        # PID controller constants for steering
        self.Kp_steering = 0.5
        self.Ki_steering = 0.3
        self.Kd_steering = 0.3

        # Initialize the PID controllers
        self.pid_steering = PID(self.Kp_steering, self.Ki_steering,
                                self.Kd_steering, setpoint=self.OVERTAKE_DISTANCE)
        self.pid_steering.output_limits = (-1, 1)  # Adjust as needed
        self.pid_steering.sample_time = 0.05

        self.testing_variable = 0

        # Memory buffer -> track change of X-value of right lane over time
        self.memory_buffer = []

    def _add_to_mem_buffer(self, x_value : float):
        if len(self.memory_buffer) >= self.MEMORY_BUFFER_LEN:
            self.memory_buffer.pop()
        self.memory_buffer.insert(0, x_value)

    def overtake(self, detection_location: str) -> Twist:
        if detection_location == DetectionLocation.FRONT or detection_location == DetectionLocation.FRONT_RIGHT:
            return self._movement.turn_left(0.5)

        elif detection_location == DetectionLocation.BACK_RIGHT or detection_location == DetectionLocation.RIGHT:
            if self._right_count < self.MAX_RIGHT_COUNT:
                return self._movement.turn_right(0.5)

        return self._movement.move_forward(2.0)

    def adjust_position(self, detection_location: str) -> Twist:
        if detection_location == DetectionLocation.BACK_RIGHT:
            if self._right_count < self.MAX_RIGHT_COUNT:
                return self._movement.turn_right(1.0)

        elif detection_location == DetectionLocation.RIGHT:
            self._right_count += 1

        elif detection_location == DetectionLocation.FRONT:
            return self._movement.turn_left(1.0)

        return self._movement.move_forward(2.0)

    def keep_dist_to_point(self, lidar_scan: dict, cluster) -> tuple[Twist, float]:
        """
        Using a PID controller return appropriate movement commands to
        stay OVERTAKE_DISTANCE from the car
        """
        # Keep the car OVERTAKE_DISTANCE away from the object
        dist_av = cluster

        # Update PID controllers with the current distance
        steering_correction = float(self.pid_steering(dist_av)) * -1

        # self.testing_variable = steering_correction

        # Implement a deadband for the steering correction
        deadband_threshold = 0.05
        if abs(steering_correction) < deadband_threshold:
            steering_correction = 0.0

        # Steer based on the PID controller output
        if steering_correction <= 0:
            return self._movement.turn_left(abs(steering_correction))
        else:
            return self._movement.turn_right(abs(steering_correction))

    def get_degree_average(self, lidar_scan: dict, key: int) -> float:
        """
        Gets average of one degree of a Lidar Scan based on specified key.
        """
        DEGREE_SIZE = 3  # Num of keys that make up 1 degree
        i = -1
        output = 0
        while i <= 1:
            addend = lidar_scan.get(key + i)
            if addend is not None:
                output = output + addend
            i += 1
        return output

    def perform_overtaking(self, msg: LaserScan, lane_x_position: float) -> tuple[Optional[Twist], Optional[float]]:
        """
        Method contains logic for overtaking movement. Method runs every lidar scan.
        Returns Twist Messages to control the car.
        """
        # Dict of detcted object bigger than 0.2 m
        temp_detected_objects = self._object_detection.detect_objects(msg)
        # Dict of all detected points (degree in keys : dist to obj)
        all_detected_objects = self._object_detection._objects_detected

        if not self._object_detection._object_detected:
            return None

        closest_cluster_width = 0
        closest_cluster_distance = 0

        if self._right_count >= 30:
            self.enough_rotations = True

        # Detect biggest cluster (expected to be cat in this case)
        loop_detected_objects = temp_detected_objects.copy()
        closest_cluster = dict()
        closest_cluster_distance = 100.0
        closest_cluster_width = 0.0
        for cluster in self._object_detection.detected_clusters:
            size, distance = self._object_detection.find_cluster_size(cluster)
            if distance < closest_cluster_distance:
                closest_cluster_distance = distance
                closest_cluster = cluster
                closest_cluster_width = size
        detection_location = self._object_detection.get_most_frequent_position(
            closest_cluster)

        # Detect a cat. Bigger 1.5 in width -> cat
        if not self.is_car_detected and closest_cluster_width >= 1.5 and closest_cluster_distance <= 5.0:
            self.is_car_detected = True

        # When car detected -> align the car to move to left lane
        if self.is_car_detected and not self.is_in_left_lane:
            if all_detected_objects.get(str(self.WAYPOINT_ANGLE)) is not None:
                # Stop turning left and start tracking dist to WAYPOINT_ANGLE
                self.is_in_left_lane = True
            return self._movement.turn_left(1.0)

        # Move forward to get into the left lane
        if self.is_car_detected and self.is_in_left_lane and not self.is_next_to_car:
            if all_detected_objects.get(str(self.NEXT_TO_CAR_ANGLE)) is not None:
                self.is_next_to_car = True
            return self._movement.move_forward(1.0)

        # Try to stay away OVERTAKE_DISTANCE from WAYPOINT_ANGLE using PID
        if self.is_car_detected and self.is_in_left_lane and self.is_next_to_car and not self.is_overtake_done:
            if all_detected_objects.get(str(self.OVERTAKE_CUTOFF_ANGLE)) is None:
                # This means kart went past the object
                self.is_overtake_done = True
            else:
                # Pass PID instructions to the Node
                return self.keep_dist_to_point(loop_detected_objects, closest_cluster_distance)

        # Merge back to right lane
        if self.is_car_detected and self.is_in_left_lane and self.is_next_to_car and self.is_overtake_done and not self.is_finished:
            # Turn right until the right degree angle sees something
            if all_detected_objects.get(str(820)) is None:
                return self._movement.turn_right(1.0)
            else:
                self.testing_variable = '810 Detected'
                self.is_finished = True
                return self._movement.move_forward(1.0)
            
        if self.is_finished:
            # GO straight until the 90 degree angle looses the cat
            if all_detected_objects.get(str(810)) is not None:
                return self._movement.move_forward(1.0)
            else:
                return None
            






