from typing import Optional

from geometry_msgs.msg import Twist
from rclpy.impl.rcutils_logger import RcutilsLogger

from core.lane_tracking.utils.camera_utils import DetectedLaneResult


class LaneSwitcher(object):
    LANE_THRESHOLD = 0.5

    def __init__(self, logger: RcutilsLogger) -> None:
        self._logger = logger

        self._twsit_message = Twist()

        # Define state
        self._switching_to_left_lane = True  # Start in the right lane

        # Has the lane switch been completed
        self._lane_switch_completed = False
        self._was_margin_passed = False

        self._previous_left_lane_dist = None
        self._previous_right_lane_dist = None

    def get_logger(self) -> RcutilsLogger:
        return self._logger

    def set_switching_to_left_lane(self, switching_to_left_lane: bool) -> None:
        if not self._lane_switch_completed:
            return

        if self._switching_to_left_lane == switching_to_left_lane:
            return

        self._lane_switch_completed = False
        self._was_margin_passed = False

        self._switching_to_left_lane = switching_to_left_lane

    def switch_lane_control_loop(self, lane_distances: DetectedLaneResult) -> Optional[Twist]:
        if self._lane_switch_completed:
            return None

        left_dist = lane_distances.get('left', {}).get('dist') if lane_distances.get(
            'left', {}
        ).get('dist') not in (0, -999) else None
        right_dist = lane_distances.get('right', {}).get('dist') if lane_distances.get(
            'right', {}
        ).get('dist') not in (0, -999) else None
        far_left_dist = lane_distances.get('far_left', {}).get('dist') if lane_distances.get(
            'far_left', {}
        ).get('dist') not in (0, -999) else None
        far_right_dist = lane_distances.get('far_right', {}).get('dist') if lane_distances.get(
            'far_right', {}
        ).get('dist') not in (0, -999) else None

        self.get_logger().info(
            f"""
                Lane distances:
                Far:     {far_left_dist} <-> {far_right_dist}
                Current: {left_dist} <-> {right_dist}
            """
        )

        if self._switching_to_left_lane:
            if left_dist is not None and far_left_dist is not None:
                if far_left_dist is None or far_left_dist > 10 or left_dist > 100 or far_right_dist is None or far_right_dist < 100:
                    self._twsit_message.linear.x = 0.5
                    self._twsit_message.angular.z = 0.5
                else:
                    self._twsit_message.linear.x = 0.5
                    self._twsit_message.angular.z = 0.0

                if self._previous_left_lane_dist is not None:
                    if self._previous_left_lane_dist - left_dist > 100:
                        self._was_margin_passed = True

                if self._was_margin_passed:
                    if left_dist <= 300:
                        self._lane_switch_completed = True

        else:
            if right_dist is not None and far_right_dist is not None:
                if far_right_dist is None or far_right_dist > 10 or right_dist > 100 or far_left_dist is None or far_left_dist < 100:
                    self._twsit_message.linear.x = 0.5
                    self._twsit_message.angular.z = -0.5
                else:
                    self._twsit_message.linear.x = 0.5
                    self._twsit_message.angular.z = 0.0

                if self._previous_right_lane_dist is not None:
                    if self._previous_right_lane_dist - right_dist > 100:
                        self._was_margin_passed = True

                if self._was_margin_passed:
                    if right_dist <= 300:
                        self._lane_switch_completed = True

        self._previous_left_lane_dist = left_dist
        self._previous_right_lane_dist = right_dist

        return self.get_twist_message()

    def get_twist_message(self) -> Twist:
        return self._twsit_message
