from typing import List, Optional, Callable
from enum import Enum
from simple_pid import PID

from rclpy.impl.rcutils_logger import RcutilsLogger
from geometry_msgs.msg import Twist

from core.utils import DetectedLidarObject
from core.lane_tracking import DetectedLaneResult


class OvertackingState(Enum):
    SWITCH_LANE_LEFT = 1
    OVERTACKE_PERPENDICULAR = 2
    SWITCH_LANE_RIGHT = 3
    OVERTAKE_FINISH = 4


class OvertackingPlanning(object):
    LONG_OBJECTS_LENGTH_THRESHOLD = 1.0  # meters
    FRONT_DISTANCE_OBJECT_THRESHOLD = 5.0  # meters
    FRONT_OBJECTS_DETECTION_COUNT = 3

    PID_MIN_OUTPUT = -5.0
    PID_MAX_OUTPUT = 5.0
    PID_DEVISION_FACTOR = 10.0

    def __init__(self, logger: RcutilsLogger, get_angle_distance: Callable[[int, Optional[int]], Optional[float]]):
        self._logger = logger
        self.get_angle_distance = get_angle_distance

        self._is_overtaking: bool = False
        self._front_object_count: int = 0
        self._twist_message: Twist = Twist()
        self._tracked_object_id: Optional[str] = None

        self._overtacke_state: OvertackingState = OvertackingState.SWITCH_LANE_LEFT

        # PID controller constants
        self.Kp = 4.0
        self.Ki = 10.5
        self.Kd = 3.2

        # Initialize the PID controller
        self.pid = PID(self.Kp, self.Ki, self.Kd, setpoint=2.5)
        self.pid.output_limits = (self.PID_MIN_OUTPUT, self.PID_MAX_OUTPUT)
        self.pid.sample_time = 0.1

    def get_logger(self):
        return self._logger

    def set_overtake_state(self, state: OvertackingState):
        self.get_logger().info(
            f"Overtake state: {state}"
        )

        self._overtacke_state = state

    def get_overtake_state(self) -> OvertackingState:
        return self._overtacke_state

    def plan(self, objects: List[DetectedLidarObject], lane_distances: DetectedLaneResult) -> Optional[Twist]:
        # Detect overtaking object
        detect_result = self._detect_overtaking_object(objects, lane_distances)

        if detect_result == True:
            # If an object was detected, turn left until the right angle is reached
            if self.get_overtake_state() == OvertackingState.SWITCH_LANE_LEFT:
                return self._turn_until_right_angle_reached(True, OvertackingState.OVERTACKE_PERPENDICULAR)

            # If the right angle is reached, overtake the object using the PID controller
            if self.get_overtake_state() == OvertackingState.OVERTACKE_PERPENDICULAR:
                tracked_object = self._get_right_tracked_object(objects)

                if tracked_object is not None:
                    return self._calculate_overtaking_twist(tracked_object)

            # If the object is overtaken, turn right until the right angle is reached
            if self.get_overtake_state() == OvertackingState.SWITCH_LANE_RIGHT:
                return self._turn_until_right_angle_reached(False, OvertackingState.OVERTAKE_FINISH, 775)

            # If the right angle is reached, move forward until the right angle is null
            if self.get_overtake_state() == OvertackingState.OVERTAKE_FINISH:
                return self._move_forward_until_right_angle_null()

        return None

    def _turn_until_right_angle_reached(self, turn_left: bool, next_state: OvertackingState, target_angle: int = 810) -> Twist:
        right_angle = self.get_angle_distance(
            target_angle - 1,
            target_angle + 1
        )
        angular_velocity = 0.5 if turn_left else -0.5

        if right_angle is None:
            return self._make_twist_message(
                angular=angular_velocity
            )
        else:
            self.set_overtake_state(next_state)

        return self._make_twist_message(0, 0)

    def _move_forward_until_right_angle_null(self) -> Twist:
        right_angle = self.get_angle_distance(819, 821)

        if right_angle is not None:
            return self._make_twist_message(0.5, 0)

        self.get_logger().info(
            "Overtaking maneuver completed"
        )

        self._is_overtaking = False
        return self._make_twist_message(0.5, 0)

    def _detect_overtaking_object(self, objects: List[DetectedLidarObject], lane_distances: DetectedLaneResult) -> bool:
        if self._is_overtaking:
            return True

        # Filter for objects that meet the length threshold
        long_objects = [
            obj for obj in objects if obj.get('length', 0) >= self.LONG_OBJECTS_LENGTH_THRESHOLD
        ]

        # Filter for objects that are in front of the vehicle
        for obj in long_objects:
            if self._is_object_in_front(obj):
                self._front_object_count += 1

        # If a front object is detected n times, then the vehicle is overtaking
        if self._front_object_count > self.FRONT_OBJECTS_DETECTION_COUNT:
            self._is_overtaking = True

            self.get_logger().info(
                f"Object in front: {obj.get('id')}"
            )

            self._front_object_count = 0
            return True

        return False

    def _is_object_in_front(self, obj: DetectedLidarObject) -> bool:
        return (
            obj.get('start_pos')[0] > 0 and
            obj.get('end_pos')[0] > 0 and
            -1.5 < self.round_off_nr(obj.get('start_pos')[1]) < 1.5 and
            -1.5 < self.round_off_nr(obj.get('end_pos')[1]) < 1.5 and
            self.get_distance_to_point(
                obj.get('closest_pos')[0], obj.get('closest_pos')[1]
            ) <= self.FRONT_DISTANCE_OBJECT_THRESHOLD
        )

    def _get_right_tracked_object(self, objects: List[DetectedLidarObject]) -> DetectedLidarObject:
        # Filter for the longest object on the right side of the kart
        right_objects = [
            obj for obj in objects if obj.get('central_pos')[1] < -1.0
        ]

        tracked_object = None

        if len(right_objects) != 0:
            tracked_object = self.get_object_by_id(
                right_objects,
                self._tracked_object_id
            )

            # If there is no object with the previous ID, track the 'new' longest object
            if tracked_object is None:
                new_right_object = max(
                    right_objects,
                    key=lambda obj: obj.get('length')
                )

                self._tracked_object_id = new_right_object.get('id')
                tracked_object = new_right_object

        if tracked_object is not None:
            self.get_logger().info(
                f"""
                    Tracked object: {tracked_object.get('id')}
                    Start: {tracked_object.get('start_pos')}
                    Central: {tracked_object.get('central_pos')}
                    End: {tracked_object.get('end_pos')}
                    Length: {tracked_object.get('length')}
                    Right Angle: {self.get_angle_distance(809, 811)}
                    Right-up Angle: {self.get_angle_distance(624, 626)}
                """
            )

        return tracked_object

    def get_object_by_id(self, objects: List[DetectedLidarObject], object_id: Optional[str]) -> Optional[DetectedLidarObject]:
        if object_id is None:
            return None

        for obj in objects:
            if obj.get('id') == object_id:
                return obj

        return None

    def _compute_pid(self, pid_input: float) -> float:
        # Perform the PID calculation
        pid_result = float(
            self.pid(pid_input)
        )

        pid_result = round(pid_result / self.PID_DEVISION_FACTOR, 2)

        self.get_logger().info(
            f"""
                PID Input: {pid_input}
                PID Result: {pid_result}
            """
        )

        return pid_result

    def _calculate_overtaking_twist(self, tracked_object: DetectedLidarObject) -> Twist:
        start_x_pos = tracked_object.get('start_pos')[0]
        end_x_pos = tracked_object.get('end_pos')[0]

        x_pos = tracked_object.get('central_pos')[0]
        y_pos = tracked_object.get('central_pos')[1]

        # By default track the distance to the right angle
        pid_input = self.get_angle_distance(800, 820)

        # If the right angle is null, and the tracked object is behind, change the state
        if start_x_pos < 0.3 and end_x_pos < 0.3:
            return self._prepare_for_right_turn()

        if pid_input is None:
            # Calculate the distance to the object
            # pid_input = self.get_distance_to_point(x_pos, y_pos)

            # Compute the PID output
            pid_input = abs(y_pos)

        # Calculate the PID output
        pid_output = self._compute_pid(pid_input)

        # Calculate the linear and angular velocities
        linear_velocity = 0.5
        angular_velocity = pid_output

        return self._make_twist_message(
            linear=linear_velocity,
            angular=angular_velocity
        )

    def _prepare_for_right_turn(self) -> Optional[Twist]:
        right_up_angle = self.get_angle_distance(750, 752)
        right_angle = self.get_angle_distance(809, 811)
        right_bottom_angle = self.get_angle_distance(819, 821)

        if right_up_angle is None and right_angle is None:
            if right_bottom_angle is None:
                self.set_overtake_state(OvertackingState.SWITCH_LANE_RIGHT)
                return None

            return self._make_twist_message(
                linear=0.5,
                angular=0.0
            )

        return None

    def _make_twist_message(self, linear: float = 0.5, angular: float = 0.0) -> Twist:
        self._twist_message.linear.x = float(linear)
        self._twist_message.angular.z = float(angular)

        return self._twist_message

    @staticmethod
    def round_off_nr(number) -> float:
        """
        Will round off the number to the nearest 0.5 value.

        :param number: The number to round off

        :return: The rounded off number
        """

        return round(number, 2) / 2

    @staticmethod
    def get_distance_to_point(point_x: float, point_y: float) -> float:
        """
        Calculates the distance to a point using the Pythagorean theorem.

        :param point_x: The x-coordinate of the point
        :param point_y: The y-coordinate of the point

        :return: The distance to the point
        """

        return (point_x**2 + point_y**2)**0.5
