from rclpy.node import Node
from sensor_msgs.msg import LaserScan
from geometry_msgs.msg import Twist
from custom_interfaces.msg import PriorityTwist

from .parallel_methods import ParallelMethods
from ..utils.object_detection import ObjectDetection

import rclpy


class ParallelParking(Node):
    def __init__(self):
        super().__init__("parallel_parking_node")
        self.get_logger().info("Parallel parking started!")

        self.subscription = self.create_subscription(
            LaserScan,
            '/scan',
            self.listener_callback,
            10
        )

        self._cmd_vel_publisher = self.create_publisher(
            PriorityTwist,
            '/priority_cmd_vel',
            10
        )
        self._priority_twist_message: PriorityTwist = PriorityTwist()

        self._vel = Twist()

        self.object_detection = ObjectDetection()
        self.parallel_methods = ParallelMethods(self.get_clock())

    def listener_callback(self, msg: LaserScan) -> None:
        self.object_detection.detect_objects(msg)
        lidar_data = self.object_detection._objects_detected

        if self.parallel_methods.is_car_ready_to_park == False:
            # Executes passing the barrier, then the car stops
            self.parallel_methods.detect_parking_spot(lidar_data)
        else:
            # Executes the parking sequence
            self.parallel_methods.parking_sequence_wrapper(
                lidar_data
            )

        # Publishes the movement of the car
        self._vel.linear.x = self.parallel_methods.current_linear_velocity
        self._vel.angular.z = self.parallel_methods.current_angular_velocity

        self.publish_velocity(self._vel)

    def publish_velocity(self, twist_message: Twist, priority: int = 4) -> None:
        self._priority_twist_message.priority = priority
        self._priority_twist_message.twist = twist_message
        self._cmd_vel_publisher.publish(self._priority_twist_message)


def main(args=None):
    rclpy.init(args=args)
    node = ParallelParking()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
