from typing import TypedDict
from rclpy.node import Node
from std_srvs.srv import SetBool
from geometry_msgs.msg import Twist
from std_msgs.msg import Bool

from custom_interfaces.srv import SetSpeedLimit, SetTrafficLight
from custom_interfaces.msg import PriorityTwist

import rclpy.time
import rclpy


class NavigationCommand(TypedDict):
    linear_velocity: float
    angular_velocity: float
    priority: int
    timestamp: rclpy.time.Time


class PlanningNode(Node):
    NAVIGATION_COMMAND_TIMEOUT = 1  # seconds

    def __init__(self):
        super().__init__('planning_node')
        self.get_logger().info('PlanningNode started')

        # Properties for determining the path
        self._can_start = False
        self._speed_limit = 10.0 / 3.6  # 10 km/h to m/s
        self._last_navigation_command: NavigationCommand = {
            'linear_velocity': 0.0,
            'angular_velocity': 0.0,
            'priority': 0,
            'timestamp': self.get_clock().now()
        }

        # Subscribe to the /priority_cmd_vel topic
        self.create_subscription(
            PriorityTwist,
            '/priority_cmd_vel',
            self._priority_cmd_vel_callback,
            10
        )

        # Subscribe to the /set_traffic_light service
        self._traffic_light_service = self.create_service(
            SetTrafficLight,
            'set_traffic_light',
            self._set_traffic_light_state_callback
        )

        # Subscribe to the /set_speed_limit service
        self._speed_limit_service = self.create_service(
            SetSpeedLimit,
            'set_speed_limit',
            self._set_speed_limit_callback
        )

        # Publish to the /cmd_vel topic
        self._cmd_vel_publisher = self.create_publisher(
            Twist,
            '/cmd_vel',
            10
        )
        self._twist_message: Twist = Twist()

        self._can_start_publisher = self.create_publisher(
            Bool,
            '/can_start',
            10
        )
        self._can_start_message: Bool = Bool()
        self._can_start_timer = self.create_timer(0.1, self._publish_can_start)

        # A service to change the _can_start property
        self._can_start_service = self.create_service(
            SetBool,
            'set_can_start',
            self._set_can_start_callback
        )

        # Create a timer to publish the last navigation command
        self.create_timer(0.1, self.start_movement)

    def get_speed_limit(self) -> float:
        return self._speed_limit

    def can_start(self) -> bool:
        return self._can_start

    def set_can_start(self, value: bool) -> None:
        self.get_logger().info(
            f'Can start changed to: {value}'
        )

        self._can_start = value

    def _publish_can_start(self) -> None:
        self._can_start_message.data = self.can_start()
        self._can_start_publisher.publish(self._can_start_message)

    def _set_navigation_command(self, linear_velocity: float, angular_velocity: float, priority: int):
        is_expired = (
            self.get_clock().now() - self._last_navigation_command.get('timestamp')
        ).nanoseconds >= self.NAVIGATION_COMMAND_TIMEOUT * 1e9  # seconds to nanoseconds

        has_higher_or_equal_priority = (
            priority <= self._last_navigation_command.get('priority')
        )

        # If the priority of the new command is higher than the last one, update the command
        if has_higher_or_equal_priority or is_expired:
            if self._last_navigation_command.get('priority') != priority:
                self.get_logger().info(
                    f"New navigation command received with priority: {priority}"
                )

            # Enforce the speed limit
            self._last_navigation_command['linear_velocity'] = min(
                linear_velocity, self.get_speed_limit()
            )
            self._last_navigation_command['angular_velocity'] = angular_velocity
            self._last_navigation_command['priority'] = priority
            self._last_navigation_command['timestamp'] = self.get_clock().now()
        else:
            # If the priority of the new command is lower than the last one, ignore the command
            # self.get_logger().info(
            #     f"Ignoring the new navigation command with priority: {priority}"
            # )
            pass

    def _priority_cmd_vel_callback(self, msg: PriorityTwist):
        # Process the PriorityTwist message
        self._set_navigation_command(
            linear_velocity=msg.twist.linear.x,
            angular_velocity=msg.twist.angular.z,
            priority=msg.priority
        )

    def _set_traffic_light_state_callback(self, request, response):
        self.get_logger().info(
            f'\n\nTraffic light state changed to: '
            f'{"GREEN" if request.is_green else "RED"}\n\n'
        )

        # Process the request and fill the response
        self.set_can_start(request.is_green)

        response.success = True
        return response

    def _set_speed_limit_callback(self, request, response):
        self.get_logger().info(
            f'\nSpeed limit changed to: {request.speed_limit}\n'
        )

        # Process the request and fill the response
        self._speed_limit = request.speed_limit
        # Update the last navigation command
        self._last_navigation_command['linear_velocity'] = min(
            self._last_navigation_command.get('linear_velocity'),
            self.get_speed_limit()
        )

        response.success = True
        return response

    def _set_can_start_callback(self, request, response):
        # Process the request and fill the response
        self.set_can_start(request.data)

        if not self._can_start:
            self.stop_movement()

        response.success = True
        return response

    def start_movement(self):
        if self.can_start():
            self._publish_velocity(
                self._last_navigation_command.get('linear_velocity'),
                self._last_navigation_command.get('angular_velocity')
            )

    def stop_movement(self):
        self.set_can_start(False)
        self._publish_velocity(0.0, 0.0)

    def _publish_velocity(self, linear_velocity: float, angular_velocity: float) -> None:
        self._twist_message.linear.x = linear_velocity
        self._twist_message.angular.z = angular_velocity
        self._cmd_vel_publisher.publish(self._twist_message)


def main(args=None):
    rclpy.init(args=args)
    node = PlanningNode()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
