from collections import Counter
from dotenv import load_dotenv
# from core.utils import display_image_safe

import numpy as np
import cv2
import os


dotenv_path = os.path.join(
    os.path.dirname(__file__),
    '.env'
)

load_dotenv(dotenv_path=dotenv_path)


class SpeedDetectionAlgorithm(object):
    SIMULATION = os.getenv('GAZEBO_SIMULATION', 'false') == 'true'
    MODEL_PATH = os.path.join(
        os.path.dirname(__file__),
        'model', 'trained_knn_model.xml'
    )

    MAX_DETECTED_SPEED = 50
    SPEED_DEVISION = 5

    def __init__(self):
        self._model = cv2.ml.KNearest_load(self.MODEL_PATH)
        self._detected_numbers: list = []
        # Threshold for the number detection; 60 for the real world and 90 for gazebo
        self._number_threshold = 90 if self.SIMULATION else 60

    # Process the image to get the number
    def _image_to_number(self, image: cv2.Mat) -> str:
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        thresh = cv2.adaptiveThreshold(gray_image, 255, 1, 1, 11, 2)
        roismall = cv2.resize(thresh, (10, 10))
        roismall = roismall.reshape((1, 100))
        roismall = np.float32(roismall)

        _, results, _, _ = self._model.findNearest(
            roismall, k=1
        )
        value = str(int((results[0][0])))
        return value

    # of 0 is at the start of the number, put it in the back
    def _move_zero_to_end(self, num_string: str) -> str:
        if num_string.startswith('0'):
            num_string = num_string[1:] + '0'
        return num_string

    # Adjust the ROI and extract the part of the image that contains a single digit
    def _detect_number(self, image: cv2.Mat) -> int:
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        gray = cv2.GaussianBlur(gray, (5, 5), 0)

        _, th = cv2.threshold(gray, self._number_threshold,
                              255, cv2.THRESH_BINARY_INV)
        contours, _ = cv2.findContours(
            th, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        numbers_in_frame = ()
        for cnt in contours:
            x, y, w, h = cv2.boundingRect(cnt)
            if (h > 70 and h < 100) and (w > 15 and w < 100):
                roi = image[y:y+h, x:x+w]
                roismall = cv2.resize(roi, (10, 10))
                numbers_in_frame = numbers_in_frame + (roismall,)

        final_number = ''
        for number in numbers_in_frame:
            final_number = final_number + self._image_to_number(number)

        final_number = self._move_zero_to_end(final_number)

        # print(final_number)
        final_number_int = int(final_number)

        if final_number_int < self.MAX_DETECTED_SPEED and len(final_number) > 1 and final_number_int % self.SPEED_DEVISION == 0:
            return final_number_int

        # The number 16 is a special case unique to the simulation
        if final_number_int == 16:
            return final_number_int

    # Mask for red color
    def _red_mask_gazebo(self, image: cv2.Mat):
        lower_red = np.array([170, 150, 50])
        upper_red = np.array([180, 255, 255])
        lower_red_extra = np.array([0, 100, 50])
        upper_red_extra = np.array([10, 255, 255])
        return cv2.inRange(image, lower_red, upper_red) + cv2.inRange(image, lower_red_extra, upper_red_extra)

    def _red_mask_real(self, image: cv2.Mat):
        lower_red = np.array([0, 50, 50])
        upper_red = np.array([10, 255, 255])
        return cv2.inRange(image, lower_red, upper_red)

    def _read_mask(self, image: cv2.Mat):
        if self.SIMULATION:
            return self._red_mask_gazebo(image)
        else:
            return self._red_mask_real(image)

    # Circle detection
    def _circle_check(self, cnt) -> bool:
        area = cv2.contourArea(cnt)
        perimeter = cv2.arcLength(cnt, True)
        circ_ratio = (4 * 3.14159 * area) / (perimeter * perimeter)

        return circ_ratio >= 0.5 and area > 700

    # Get the sign(circle) ROI
    def _get_circle_roi(self, image: cv2.Mat, cnt):
        x, y, w, h = cv2.boundingRect(cnt)
        circle_roi = image[y:y+h, x:x+w]
        circle_roi = cv2.resize(circle_roi, (200, 200))
        # display_image_safe(circle_roi, 'circle_roi')

        # check if the circle is in the right part of the image
        width = image.shape[1]
        right_most_boundary = (2 * width) // 3

        return circle_roi, x >= right_most_boundary

    # Process the whole frame to detect the sign
    def _sign_detection(self, main_image: cv2.Mat) -> int:
        img_hsv = cv2.cvtColor(main_image, cv2.COLOR_BGR2HSV)
        mask = self._read_mask(img_hsv)

        # display_image_safe(main_image, 'image')

        contours, _ = cv2.findContours(
            mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
        )

        for cnt in contours:
            try:
                if self._circle_check(cnt):
                    circle_roi, is_right = self._get_circle_roi(
                        main_image, cnt
                    )
                    if not is_right:
                        # display_image_safe(circle_roi, 'circle_roi'
                        self._detected_numbers.append(
                            self._detect_number(circle_roi)
                        )
                        return None
                    else:
                        counter = Counter(self._detected_numbers)
                        most_common_num = counter.most_common(2)
                        if most_common_num[0][0] is not None:
                            return most_common_num[0][0]
                        else:
                            return most_common_num[1][0]
            except:
                continue

    def detect_speed_from_image(self, image: cv2.Mat) -> int | None:
        width = image.shape[1]
        split_width = width // 6
        right_half = image[:, 3*split_width:]
        number = self._sign_detection(right_half)

        if number is not None:
            self._detected_numbers = []

        # display_image_safe(right_half, 'right_half')

        return number


if __name__ == '__main__':
    detect_alg = SpeedDetectionAlgorithm()
    # imaage = '/home/miroo/RDW-Stuff/maveric-dev/rdw-maveric/maverick_ws/src/core/core/speed/Screenshot 2024-05-09 183943.png'
    # number = _detect_alg.detect_number_from_specific_image(cv2.imread(imaage))
    # if number:
    #     print(number)
    # cv2.waitKey(0)

    # pathed = "/home/miroo/RDW-Stuff/rdw-maveric/TestData/Video_Short.webm"
    # path = '/home/miroo/RDW-Stuff/extra_stuff/front_short'

    # from perception.perception.utils import get_camera_config
    # To get the front camera use `get_camera_config('front')`

    cap = cv2.VideoCapture(0)
    while cap.isOpened():

        ret, frame = cap.read()

        number = detect_alg.detect_speed_from_image(frame)
        if number:
            print("The real one:" + str(number))
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break
    cap.release()

    # for filename in sorted(os.listdir(path)):
    #     img_path = os.path.join(path, filename)
    #     image = cv2.imread(img_path)
    #     width = image.shape[1]
    #     split_width = width // 6
    #     right_half = image[:, 3*split_width:]
    #     number = _detect_alg.detect_number_from_specific_image(right_half)
    #     if number != None:
    #         global detected_numbers
    #         detected_numbers = []
    #         print(number)
