from sensor_msgs.msg import Image
from rclpy.node import Node

from custom_interfaces.srv import SetSpeedLimit
from cv_bridge import CvBridge

from .detect_speed import SpeedDetectionAlgorithm
from core.utils import display_image_safe

import rclpy
import os


class SpeedLimitNode(Node):
    def __init__(self):
        super().__init__('speed_limit_node')
        self.get_logger().info('SpeedLimitNode started')

        # Subscribe to the /camera_front/image_raw topic
        self.create_subscription(
            Image,
            '/camera_front/image_raw',
            self._image_front_callback,
            10
        )
        self._bridge = CvBridge()

        self._client = self.create_client(SetSpeedLimit, 'set_speed_limit')
        while not self._client.wait_for_service(timeout_sec=1.0):
            self.get_logger().info('service not available, waiting again...')

        self._request = SetSpeedLimit.Request()

        # Speed Detection Algorithm
        self._detect_alg = SpeedDetectionAlgorithm()

        # Check if the simulation environment variable is set
        if os.getenv('GAZEBO_SIMULATION') is None:
            self.get_logger().warn(
                'SIMULATION environment variable is not set. Using default value of false'
            )

    def _image_front_callback(self, msg: Image) -> None:
        # Convert the ROS Image message to OpenCV image
        image = self._bridge.imgmsg_to_cv2(msg, 'bgr8')
        # display_image_safe(image, 'Front Camera')

        number = self._detect_alg.detect_speed_from_image(image)

        if number is not None:
            self.get_logger().info(f'Detected speed limit: {number}')
            # Convert from km/h to m/s
            number = float(number / 3.6)
            self.send_speed_limit_request(number)

    def send_speed_limit_request(self, speed_limit: float) -> None:
        self._request.speed_limit = speed_limit
        self._future = self._client.call_async(self._request)


def main(args=None):
    rclpy.init(args=args)

    node = SpeedLimitNode()

    rclpy.spin(node)

    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
