from enum import Enum
from ultralytics import YOLO
# from core.utils import display_image_safe

import numpy as np
import cv2
import os


class TrafficLightColor(Enum):
    RED = 'red'
    GREEN = 'green'


class TrafficLightDetection(object):
    # Constant values
    AREA_THRESHOLD = 75
    MAX_AREA = 900
    CIRC_RATIO_THRESHOLD = 0.5
    ROI_BOX_SIZE = 300
    MODEL_PATH = os.path.join(
        os.path.dirname(__file__),
        'model', 'yolov8n.pt'
    )

    def __init__(self):
        self._model = YOLO(self.MODEL_PATH)
        self._red_counter, self._green_counter, self.emergency_counter = 0, 0, 0

    # we only care about the right side of the image as the traffic light is always on the right
    def _get_right_side(self, image):
        # Get the width of the image
        width = image.shape[1]

        # Calculate the starting and ending columns for the right half
        start_col = width // 2
        end_col = width

        # Extract the right half of the image
        right_half = image[:, start_col:end_col]

        return right_half

    def _get_green_mask(self, hsv_img):
        low = np.array([45, 100, 100])
        up = np.array([100, 255, 255])
        return cv2.inRange(hsv_img, low, up)

    def _get_red_mask(self, hsv_img):
        red_lower = np.array([160, 100, 100], np.uint8)
        red_upper = np.array([179, 255, 255], np.uint8)
        return cv2.inRange(hsv_img, red_lower, red_upper)

    # Get coordinates of a traffic light circle based on mask input. Finds coords of supposed stop light (green or red)
    # TODO: Remove the unused variable 'image'
    def _get_traffic_light_circle_coords(self, mask, image) -> list:
        # Get contours and their relations (hierarchy)
        contours, _ = cv2.findContours(
            mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        output_coords = []
        centroid_coord = None
        for cnt in contours:
            # Get circularity index (0-1.0) how round is a contour
            area = cv2.contourArea(cnt)
            perimeter = cv2.arcLength(cnt, True)
            if perimeter != 0:
                # circ_ratio = (4 * 3.14159 * area) / (perimeter * perimeter)
                # Eliminate trash -> assume that a light is like a circle and has a sizable area
                if self.AREA_THRESHOLD < area < self.MAX_AREA:
                    # print(area)
                    # cv2.drawContours(image, [cnt], 0, (0, 255, 0), 2)
                    # cv2.imshow("Contours", image)
                    # cv2.waitKey(0)
                    # Get coords of light
                    M = cv2.moments(cnt)  # ways to describe the contour
                    cx = int(M['m10'] / M['m00'])  # Get centroid x coord
                    cy = int(M['m01'] / M['m00'])  # Get centroid y coord
                    output_coords.append((cx, cy, area))

        # Filter found circles. Find the highest-located point if there is more than one detected
        num_of_coord = len(output_coords)
        if num_of_coord > 1:
            top_cord = output_coords[0]
            for coord in output_coords:
                if coord[1] < top_cord[1]:
                    top_cord = coord
            centroid_coord = top_cord
        elif num_of_coord == 1:
            centroid_coord = output_coords[0]

        # Output is tuple like (cx,cy,area)
        return centroid_coord

    # RG stands for red, green
    def _get_rg_coords_from_img(self, image):

        # Scale image down using 1:3 scale. shape[0] - height, [1] - width
        # convert to HSV colors
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        mask_red = self._get_red_mask(hsv)  # Get red mask
        mask_green = self._get_green_mask(hsv)  # Get red mask

        # display_image_safe(mask_red, 'RedMask')
        # display_image_safe(mask_green, 'GreenMaks')
        # cv2.imshow("GreenMask", mask_green)
        # cv2.imshow("RedMask", mask_red)

        # Get potential stoplight lights coordinates
        green_coords = self._get_traffic_light_circle_coords(mask_green, image)
        red_coords = self._get_traffic_light_circle_coords(mask_red, image)
        return green_coords, red_coords  # red_coords

    # extract a region of interest around a detected stop light
    def _extract_box(self, image, coords):
        half_box_size = self.ROI_BOX_SIZE // 2
        cx = coords[0]
        cy = coords[1]
        top_left = (max(0, cx - half_box_size), max(0, cy - half_box_size))
        bottom_right = (min(image.shape[1] - 1, cx + half_box_size),
                        min(image.shape[0] - 1, cy + half_box_size))
        return image[top_left[1]:bottom_right[1], top_left[0]:bottom_right[0]]

    # If found, return the coordinates of a centroid of a traffic light
    def get_traffic_light_color(self, image_raw: cv2.Mat) -> TrafficLightColor | None:
        # Signs are only on the right of the road
        image = self._get_right_side(image_raw)
        # image = cv2.resize(
        #     image, (int(image.shape[1] * 0.3), int(image.shape[0] * 0.3))
        # )
        # display_image_safe(image, 'Right Side')
        # cv2.imshow("Right Side", image)

        green_coords, red_coords = self._get_rg_coords_from_img(image)
        print(green_coords, red_coords)
        # If coordinates are not null use a CNN to check if the region is a Stop Light
        
        if green_coords is not None:
            self._green_counter += 1
            if self._green_counter == 5:
                self._red_counter, self._green_counter = 0, 0
                roi_green = self._extract_box(image, green_coords)
                # display_image_safe(roi_green, 'Green')
                # cv2.imshow("roiGreen", roi_green)
                results = self._model.predict(roi_green, verbose=False)
                if results and results[0].boxes and results[0].boxes[0].cls:
                    element = results[0].boxes[0].cls[0]
                    if 9 in element:
                        return TrafficLightColor.GREEN

        if red_coords is not None:
            self._red_counter += 1
            if self._red_counter == 15:
                self._red_counter, self._green_counter = 0, 0
                roi_red = self._extract_box(image, red_coords)
                # cv2.imshow("roiRed", roi_red)
                results = self._model.predict(roi_red, verbose=False)
                if results and results[0].boxes and results[0].boxes[0].cls:
                    element = results[0].boxes[0].cls[0]
                    if 9 in element:
                        return TrafficLightColor.RED

        if red_coords is None:
            print("Emergency")
            self.emergency_counter += 1
            if self.emergency_counter == 1:
                results = self._model.predict(image, verbose=False)
                if results and results[0].boxes and results[0].boxes[0].cls:
                    element = results[0].boxes[0].cls[0]
                    if 9 in element:
                        return TrafficLightColor.GREEN

        print(self._green_counter, self._red_counter)
        return None


if __name__ == '__main__':
    traffic_light_detection = TrafficLightDetection()

    video_path = "/home/miroo/RDW-Stuff/rdw-maveric/TestData/Video_Short.webm"
    image_folder_path = '/home/miroo/RDW-Stuff/extra_stuff/front_short'

    # from perception.utils import get_camera_config
    # To get the front camera use `get_camera_config('front')`

    # camera_path = get_camera_config('front')
    # camera_path = '/dev/v4l/by-path/pci-0000:00:14.0-usb-0:5.3:1Q.0-video-index0'
    # camera_path = '/dev/video4'

    # cap = cv2.VideoCapture(video_path)
    # while cap.isOpened():
    #     ret, frame = cap.read()
    #     print(traffic_light_detection.get_traffic_light_color(frame))
    #     if cv2.waitKey(1) & 0xFF == ord('q'):
    #         break
    # cv2.destroyAllWindows()
    # cap.release()

    # for filename in sorted(os.listdir(image_folder_path)):
    #     img_path = os.path.join(image_folder_path, filename)
    #     image = cv2.imread(img_path)
    #     traffic_light_detection.get_traffic_light_color(image)
    #     if cv2.waitKey(1) & 0xFF == ord('q'):
    #         break
    # cv2.destroyAllWindows()

    # GOOD
    image_green = '/home/miroo/RDW-Stuff/extra_stuff/traffic_light/2024-05-17-154716 (3).jpg'
    # GOOD
    image_green_final = '/home/miroo/RDW-Stuff/extra_stuff/traffic_light/1717146518.979809.png'
    # GOOD
    image_new_green = '/home/miroo/RDW-Stuff/extra_stuff/traffic_light/WIN_20240524_10_06_26_Pro.jpg'
    # NO GOOD
    image_new_green_two = '/home/miroo/RDW-Stuff/extra_stuff/traffic_light/WIN_20240524_10_06_35_Pro (2).jpg'
    # GOOD
    image_red = '/home/miroo/RDW-Stuff/extra_stuff/traffic_light/2024-05-17-154736 (4).jpg'
    print(traffic_light_detection.get_traffic_light_color(cv2.imread(image_green_final)))
    if cv2.waitKey(0) & 0xFF == ord('q'):
        cv2.destroyAllWindows()
