from rclpy.node import Node

from sensor_msgs.msg import Image
from std_msgs.msg import String
from cv_bridge import CvBridge

from .light_detection import TrafficLightColor, TrafficLightDetection
# from ..utils import display_image_safe
from ..utils import HorizontalLineSubscriber

import rclpy


class TrafficLightNode(Node, HorizontalLineSubscriber):
    LINE_TYPE = 'horizontal'

    def __init__(self):
        Node.__init__(self, 'traffic_light_node')
        HorizontalLineSubscriber.__init__(
            self,
            clock=self.get_clock(),
            create_client=self.create_client,
            create_timer=self.create_timer,
            logger=self.get_logger()
        )
        self.get_logger().info('TrafficLightNode started')

        # Subscribe to the /camera_front/image_raw topic
        self.create_subscription(
            Image,
            '/camera_front/image_raw',
            self._image_front_callback,
            10
        )

        # Subscribe to the /horizontal_line_detect topic
        self.create_subscription(
            String,
            '/horizontal_line_detect',
            self._horizontal_line_detect_callback,
            10
        )

        self._bridge = CvBridge()
        self._detect_alg = TrafficLightDetection()

    def _image_front_callback(self, msg: Image) -> None:
        # Return if the node is not 'working'
        if not self.is_working():
            return

        # Process the Image message
        image = self._bridge.imgmsg_to_cv2(msg, 'bgr8')

        color = self._detect_alg.get_traffic_light_color(image)

        if color is not None:
            self.get_logger().info(f'Detected traffic light: {color.value}')
            self.send_traffic_light_request(color)

    def send_traffic_light_request(self, traffic_light_color: TrafficLightColor) -> None:
        can_drive = False

        if traffic_light_color == TrafficLightColor.GREEN:
            can_drive = True

        self.send_kart_move_request(can_drive)


def main(args=None):
    rclpy.init(args=args)

    node = TrafficLightNode()

    rclpy.spin(node)

    node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
