from typing import Callable

from rclpy.clock import Clock
from rclpy.node import Client, Timer
from rclpy.impl.rcutils_logger import RcutilsLogger
from std_srvs.srv import SetBool
from std_msgs.msg import String

import time


class HorizontalLineSubscriber(object):
    COOLDOWN_TIME = 5  # seconds
    WAIT_TIME = 2  # seconds
    RESET_CHECKUP_TIME = 3  # seconds
    LINE_TYPE = 'none'

    def __init__(
            self,
            clock: Clock,
            create_client: Callable[[], Client],
            create_timer: Callable[[], Timer],
            logger: RcutilsLogger
    ):
        self._clock = clock
        self._logger = logger

        self._is_working = False
        self._was_started = False
        self._time = clock.now()

        # Make a timer to reset the is_working flag
        create_timer(
            self.RESET_CHECKUP_TIME,
            self._reset_is_working
        )

        # Service to stop the kart
        self._client = create_client(SetBool, 'set_can_start')
        while not self._client.wait_for_service(timeout_sec=1.0):
            logger.info('service not available, waiting again...')

        self._stop_request = SetBool.Request()

    def _reset_is_working(self) -> None:
        if self.is_working() and (self._clock.now() - self._time).nanoseconds >= self.COOLDOWN_TIME * 1e9:
            self.set_is_working(False)

    def send_kart_move_request(self, request: bool) -> None:
        if request == self._was_started:
            return

        self._was_started = request
        self._stop_request.data = request
        self._future = self._client.call_async(self._stop_request)

    def _horizontal_line_detect_callback(self, msg: String) -> None:
        if msg.data == self.LINE_TYPE:
            # Wait for a few seconds before starting the pedestrian crossing
            time.sleep(self.WAIT_TIME)
            # Start the pedestrian crossing
            self.set_is_working(True)
            self.set_was_started(False)
            self.set_time(self._clock.now())

    def is_working(self) -> bool:
        return self._is_working

    def set_is_working(self, is_working: bool) -> None:
        self._is_working = is_working

    def get_time(self) -> Clock:
        return self._time

    def set_time(self, time: Clock) -> None:
        self._time = time

    def was_started(self) -> bool:
        return self._was_started

    def set_was_started(self, was_started: bool) -> None:
        self._was_started = was_started
