from typing import List, TypedDict, Optional

from sensor_msgs.msg import LaserScan
from sklearn.cluster import DBSCAN
from scipy.spatial.distance import pdist

import numpy as np
import uuid
# import pprint

# import matplotlib
# import matplotlib.pyplot as plt
# matplotlib.use('TkAgg')


class DetectedLidarObject(TypedDict):
    id: str
    start_pos: np.ndarray
    central_pos: np.ndarray
    end_pos: np.ndarray
    closest_pos: np.ndarray
    furthest_pos: np.ndarray
    length: float


class LidarProcessor(object):
    IGNORE_THRESHOLD = 1.0  # meters
    TRACKING_DISTANCE_THRESHOLD = 1.0  # meters

    def __init__(self):
        # Detected objects from the most recent scan
        self._detected_objects: List[DetectedLidarObject] = []
        self._filtered_detected_objects: List[DetectedLidarObject] = []

        # Used to track objects across multiple scans
        self._tracked_objects: List[DetectedLidarObject] = []

        # Important angles
        self._ranges = 0

    def __new__(cls, *args, **kwargs):
        # Will only create a new instance if one does not already exist (singleton pattern)
        if not hasattr(cls, "_instance"):
            cls._instance = super(LidarProcessor, cls).__new__(cls)

        return cls._instance

    def get_angle_distance(self, first_angle: int = 809, second_angle: Optional[int] = 811) -> Optional[float]:
        result = 0.0

        if second_angle is not None:
            result = np.mean(self._ranges[first_angle:second_angle])
        else:
            result = self._ranges[first_angle]

        return result if np.isfinite(result) else None

    def get_filtered_detected_objects(self) -> List[DetectedLidarObject]:
        self._filter_detected_objects(self._detected_objects)
        return self._filtered_detected_objects

    def process_scan(self, scan: LaserScan) -> List[DetectedLidarObject]:
        """
        Processes the LaserScan message to detect objects and their properties.

        :param scan: The LaserScan message to process
        """
        ranges = np.array(scan.ranges)
        angles = np.linspace(scan.angle_min, scan.angle_max, len(ranges))

        # Get the right angle (90 degree) distance (average from 809 till 811)
        self._ranges = ranges

        # Remove NaNs and Inf values
        valid_indices = np.isfinite(ranges)
        ranges = ranges[valid_indices]
        angles = angles[valid_indices]

        # Convert polar coordinates to Cartesian coordinates
        x = ranges * np.cos(angles)
        y = ranges * np.sin(angles)

        # Combine x and y into a list of points
        points = np.vstack((x, y)).T

        # Cluster points to identify objects
        clusters = self._cluster_points(points)

        # Process clusters to determine object positions and lengths
        objects = [
            self._calculate_object_properties(cluster) for cluster in clusters
        ]

        # Store and filter the detected objects
        self._detected_objects = objects

        return self.get_filtered_detected_objects()

    def _cluster_points(self, points: np.ndarray, distance_threshold: float = 0.5) -> List[np.ndarray]:
        """
        Clusters points using the DBSCAN algorithm to identify potential objects.

        :param points: A list of points in Cartesian coordinates
        :param distance_threshold: The maximum distance between two points to be considered in the same cluster

        :return: A list of clusters, where each cluster is a list of points
        """
        # Use DBSCAN to cluster points
        clustering = DBSCAN(eps=distance_threshold, min_samples=2).fit(points)
        labels = clustering.labels_

        # Extract clusters
        unique_labels = set(labels)
        clusters = []
        for label in unique_labels:
            if label == -1:  # Noise
                continue

            cluster = points[labels == label]
            clusters.append(cluster)

        return clusters

    def _calculate_object_properties(self, cluster: np.ndarray) -> DetectedLidarObject:
        """
        Calculates the properties of an object, including its position and length.

        :param cluster: A cluster of points representing an object
        :return: The position (centroid) of the object and its length.
        """
        # Calculate the centroid of the cluster
        central_pos = np.mean(cluster, axis=0)

        # Calculate useful positions
        start_pos = cluster[0]
        end_pos = cluster[-1]
        furthest_pos = cluster[np.argmax(cluster[:, 0])]
        closest_pos = cluster[np.argmin(cluster[:, 0])]

        # Calculate the length (distance between start and end points)
        # length = np.linalg.norm(end_pos - start_pos)

        # Calculate the length (max distance between points in the cluster)
        length = np.max(pdist(cluster))

        # Make and return the detected object
        detected_object: DetectedLidarObject = {
            "start_pos": start_pos.round(4),
            "central_pos": central_pos.round(4),
            "end_pos": end_pos.round(4),
            "closest_pos": closest_pos.round(4),
            "furthest_pos": furthest_pos.round(4),
            "length": round(length, 2)
        }

        return detected_object

    def track_detected_objects(self, new_objects: List[DetectedLidarObject]) -> List[DetectedLidarObject]:
        """
        Updates the tracked objects with new detections and assigns unique identifiers.

        :param new_objects: A list of new detected objects
        :return: A list of tracked objects with unique identifiers
        """

        updated_objects: List[DetectedLidarObject] = []

        for new_object in new_objects:
            start_pos = new_object.get("start_pos")
            # end_pos = new_object.get("end_pos")
            length = new_object.get("length")
            object_found = False

            for old_object in self._tracked_objects:
                old_object_id = old_object.get("id")
                old_start_pos = old_object.get("start_pos")
                old_length = old_object.get("length")

                if (
                    np.linalg.norm(
                        start_pos - old_start_pos
                    ) < self.TRACKING_DISTANCE_THRESHOLD
                ):
                    new_object["id"] = old_object_id
                    updated_objects.append(new_object)
                    object_found = True
                    break

            if not object_found:
                new_id = str(uuid.uuid4())
                new_object["id"] = new_id
                updated_objects.append(new_object)

        self._tracked_objects = updated_objects

        return self._tracked_objects

    def _filter_detected_objects(self, objects: List[DetectedLidarObject]) -> List[DetectedLidarObject]:
        # Filter out objects that are too close to the LIDAR
        detected_objects = []

        for detected_object in objects:
            central_pos_x = abs(detected_object.get("central_pos")[0])
            central_pos_y = abs(detected_object.get("central_pos")[1])

            if (
                central_pos_x < self.IGNORE_THRESHOLD and
                central_pos_y < self.IGNORE_THRESHOLD
            ):
                continue

            detected_objects.append(detected_object)

        self._filtered_detected_objects = detected_objects

        return detected_objects
