from enum import Enum
from sensor_msgs.msg import LaserScan

import math


class DetectionLocation(Enum):
    FRONT = 'front'
    FRONT_RIGHT = 'frontright'
    RIGHT = 'right'
    BACK_RIGHT = 'backright'
    BACK_LEFT = 'backleft'
    LEFT = 'left'
    FRONT_LEFT = 'frontleft'
    NONE = 'none'


class ObjectDetection(object):
    # detection parameters
    CHAIR_LOWER_BOUND = 53 # in the simulation the start of the chair is detected at 53 degrees
    CHAIR_UPPER_BOUND = 1026 # in the simulation the end of the chair is detected at 1026 degrees
    DISTANCE_THRESHOLD = 5 # objects will only be detected within this distance in meters (5 meters)
    SMALL_OBJECT_THRESHOLD = 0.2 # objects with a width less than this value will be filtered out (0.2 meters)

    # position parameters
    FRONT_START = 473
    FRONT_END = 608
    FRONT_RIGHT_START = 540
    FRONT_RIGHT_END = 675
    RIGHT_START = 675
    RIGHT_END = 945
    BACK_RIGHT_START = 870
    BACK_RIGHT_END = 1079
    BACK_LEFT_START = 0
    BACK_LEFT_END = 135
    LEFT_START = 135
    LEFT_END = 405
    FRONT_LEFT_START = 405
    FRONT_LEFT_END = 540


    def __init__(self):
        self._objects_detected = dict() # dict of object detected without filtering small objects
        self._object_detected = False   # boolean if an object is detected in the front of the kart. Will only turn back to false detectionlocation is set to NONE
        self._detection_location = DetectionLocation.NONE # enum of the rough location of the detected object
        self.detected_clusters = list() # list of clusters of objects detected (list contains multiple dictionaries)
        self._cartesian_coordinates = list() # list of cartesian coordinates of the detected objects

    def detect_objects(self, msg: LaserScan) -> dict:
        objects_detected = dict()  # dict of objects detected
        self._cartesian_coordinates = self.convert_to_carteasian(msg)
        count = 0
        for value in msg.ranges:
            # exclude the chair and other kart related objects that should not be detected
            if count <= self.CHAIR_LOWER_BOUND or count >= self.CHAIR_UPPER_BOUND:
                count += 1
                continue
            if value < self.DISTANCE_THRESHOLD:  # if an object is detected within the distance
                # add the object to the list
                objects_detected[str(count)] = value
            count += 1  # increment the count

        # find clusters of objects detected and put them in the publically accessible variable
        self.detected_clusters = self.find_clusters(objects_detected)
        # Set the publically accessible objects_detected variable to all objects
        self._objects_detected = objects_detected
        
        # filter out small objects
        temp_clusters = dict()
        for cluster in self.detected_clusters:
            width, _ = self.find_cluster_size(cluster)
            if width > self.SMALL_OBJECT_THRESHOLD:
                temp_clusters.update(cluster)
        objects_detected = temp_clusters
        # find the most frequent position of the detected objects
        self._detection_location = self.get_most_frequent_position(
            objects_detected)
        
        if self._detection_location == DetectionLocation.FRONT:
            self._object_detected = True
        if self._detection_location == DetectionLocation.NONE:
            self._object_detected = False

        return objects_detected

    def get_most_frequent_position(self, objects_detected: dict) -> DetectionLocation:
        front_count = 0
        front_right_count = 0
        right_count = 0
        back_right_count = 0
        back_left_count = 0
        left_count = 0
        front_left_count = 0
        for key in objects_detected.keys():
            if int(key) >= self.BACK_LEFT_START and int(key) <= self.BACK_LEFT_END:
                back_left_count += 1
            if int(key) >= self.LEFT_START and int(key) <= self.LEFT_END:
                left_count += 1
            if int(key) >= self.FRONT_LEFT_START and int(key) <= self.FRONT_LEFT_END:
                front_left_count += 1
            if int(key) >= self.FRONT_START and int(key) <= self.FRONT_END:
                front_count += 1
            if int(key) >= self.FRONT_RIGHT_START and int(key) <= self.FRONT_RIGHT_END:
                front_right_count += 1
            if int(key) >= self.RIGHT_START and int(key) <= self.RIGHT_END:
                right_count += 1
            if int(key) >= self.BACK_RIGHT_START and int(key) <= self.BACK_RIGHT_END:
                back_right_count += 1

        max_count = max(
            front_count, front_right_count, right_count,
            back_right_count, back_left_count, left_count, front_left_count
        )

        if max_count == 0:
            return DetectionLocation.NONE
        elif max_count == front_count:
            return DetectionLocation.FRONT
        elif max_count == front_right_count:
            return DetectionLocation.FRONT_RIGHT
        elif max_count == right_count:
            return DetectionLocation.RIGHT
        elif max_count == back_right_count:
            return DetectionLocation.BACK_RIGHT
        elif max_count == back_left_count:
            return DetectionLocation.BACK_LEFT
        elif max_count == left_count:
            return DetectionLocation.LEFT
        elif max_count == front_left_count:
            return DetectionLocation.FRONT_LEFT

    def find_clusters(self, objects_detected: dict) -> list:
        clusters = list()
        cluster_count = 0
        current_number = 0
        previous_number = 0
        cluster = dict()
        for key in objects_detected.keys():
            current_number = int(key)
            if current_number - previous_number > 1:
                clusters.append(cluster)
                cluster.clear()
                cluster_count += 1
            cluster[key] = objects_detected[key]
            previous_number = current_number

        return clusters

    def find_cluster_center(self, cluster: dict) -> int:
        center = math.floor(cluster[0] - cluster[len(cluster) - 1] / 2)
        return center
    
    def find_middle_distance(self, a, b, c) -> float:
        if c == 0:
            return 0
        # semi perimeter
        s = (a + b + c) / 2
        # area of triangle
        area = math.sqrt(s * (s - a) * (s - b) * (s - c))
        # height of the triangle
        return 2 * area / c

    def find_cluster_size(self, cluster: dict) -> tuple[float, float]:
        start_distance = 0
        end_distance = 0
        start_degree = 0
        end_degree = 0
        shortest_distance = 0.0
        loop = False
        for key, value in cluster.items():
            if loop == False:
                start_degree = int(key)/3
                start_distance = value
                loop = True
                shortest_distance = value
            end_degree = int(key)/3
            end_distance = value
            if value < shortest_distance:
                shortest_distance = value

        angle_radians = math.radians(end_degree - start_degree)

        # Law of cosines
        cluster_size = math.sqrt(
            (math.pow(start_distance, 2) + math.pow(end_distance, 2)) -
            (2 * start_distance * end_distance * math.cos(angle_radians))
        )
        average_dist = self.find_middle_distance(start_distance, end_distance, cluster_size)
        return cluster_size, average_dist

    def write_values_to_file(self, msg: LaserScan) -> None:
        file = open('testfile.txt', 'a')
        text = ""
        count = 0
        for value in msg.ranges:
            if count == 620:
                file.write("620: " + str(value) + "\n")
            count += 1

        file.write("\n")
        file.write("All values: " + text)
        # file.write(str(len(msg.ranges))) # make sure we get 1080 values
        file.close()

    def convert_to_carteasian(self, msg: LaserScan) -> list:
        cartesian_coordinates = list()
        count = 0
        for value in msg.ranges:
            x = value * math.cos(math.radians(count))
            y = value * math.sin(math.radians(count))
            cartesian_coordinates.append((x, y))
            count += 1
        return cartesian_coordinates