from rclpy.node import Node
from sensor_msgs.msg import Image

from ..utils.camera_conf import get_camera_config
from ..utils.camera_conf import CamConfig

from core.utils import display_image_safe

import threading
import cv2


class BaseCameraNode(Node):
    def __init__(self, camera_name: str):
        # Validate the camera name
        if camera_name not in CamConfig.__annotations__:
            raise ValueError(f'Invalid camera name: {camera_name}')

        super().__init__(f'camera_{camera_name}_node')
        self.publisher_ = self.create_publisher(
            Image,
            f'/camera_{camera_name}/image_raw',
            10
        )

        # Initialize your camera here
        path = get_camera_config().get(camera_name)
        self.get_logger().info(f'Path: {path}')
        self._camera_capture = cv2.VideoCapture(path)

        # Start a new thread to capture images
        self._is_thread_running = True
        self._capture_thread = threading.Thread(target=self._capture_images)
        self._capture_thread.start()

    def _capture_images(self):
        # Loop to continuously capture and publish images
        while self._is_thread_running:
            # Capture image from camera
            image = self._get_next_frame()
            # Publish the image
            if image is not None:
                self.publisher_.publish(image)

    def _get_next_frame(self) -> Image | None:
        # Read the image from the camera
        _, image = self._camera_capture.read()

        if image is None:
            return None

        # Display the image
        # display_image_safe(image, self.get_name())

        # Convert the image to ROS Image message
        msg = Image()
        msg.data = image.tobytes()
        msg.width = image.shape[1]
        msg.height = image.shape[0]
        msg.encoding = 'bgr8'
        msg.step = image.shape[1] * 3

        return msg

    def destroy_node(self):
        # Stop the thread
        self._is_thread_running = False
        self._capture_thread.join()

        return super().destroy_node()
